# -*- mode: python ; coding: utf-8 -*-
#
# Python-slip39 -- Ethereum SLIP-39 Account Generation and Recovery
#
# Copyright (c) 2022, Dominion Research & Development Corp.
#
# Python-slip39 is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.  It is also available under alternative (eg. Commercial) licenses, at
# your option.  See the LICENSE file at the top of the source tree.
#
# Python-slip39 is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#

from __future__		import annotations

import cmath
import codecs
import logging
import math

try:
    from numpy		import fft as np_fft
except ImportError:
    np_fft			= None

from collections	import namedtuple, defaultdict
from typing		import List, Union, Tuple, Optional, Callable, Sequence

from ..util		import mixed_fraction, ordinal, commas, is_power_of_2, avg, rms

log				= logging.getLogger( __package__ )


def fft( x ):
    """Computes frequency bin amplitude (real) and phase (imaginary) for N/2 frequency bins
    0,..,Nyquist, for an N-sampled signal.  Uses numpy.fft.fft if available, otherwise, pfft where N
    is a power of 2, dft for other N.

    """
    if np_fft:
        return [ complex(b) for b in np_fft.fft( x ) ]
    N				= len( x )
    if not is_power_of_2( N ):
        return dft( x )
    return pfft( x )


def ifft( y: List[complex] ) -> List[complex]:
    """Compute inverse FFT for any complex bins N (ideally a power of 2, or multiple of 2."""
    if np_fft:
        return [ complex(s) for s in np_fft.ifft( y ) ]
    return idft( y )


def pfft( x ):
    """A simple pure-python FFT"""
    N				= len( x )
    if N <= 1:
        return x
    even			= pfft( x[0::2] )
    odd				= pfft( x[1::2] )
    T				= [ cmath.exp( -2j * cmath.pi * k / N ) * odd[k] for k in range( N//2 ) ]
    return [even[k] + T[k] for k in range(N//2)] + \
           [even[k] - T[k] for k in range(N//2)]


def dft( x: List[Union[int, float, complex]] ) -> List[complex]:
    """Takes N either real or complex signal samples, yields complex DFT bins.  Assuming the input
    waveform is filtered to only contain signals in the bandwidth B range -B/2:+B/2 around baseband
    frequency MID, and is frequency shifted (divided by) your baseband frequency MID, and is sampled
    at the Nyquist rate R: given N samples, the result contains N signal frequency component bins:

        index:      0                 N/2-1      N/2      N/2+1       N-1
        baseband:  [MID+] [MID+] ... [MID+]    [MID+/-]  [MID+]  ... [MID+]
        frequency:  DC     1B/N   (N/2-1)B/N   (N/2)B/N  (1-N/2)B/N  -1B/N

    """
    N				= len( x )
    result			= []
    for k in range( N ):
        r			= 0
        for n in range( N ):
            t			= -2j * cmath.pi * k * n / N
            r		       += x[n] * cmath.exp( t )
        result.append( r )
    return result


def idft( y: List[complex] ) -> List[complex]:
    """Inverse DFT on complex frequency bins.  Very slow O(N^2), so use ... carefully.

    Since we're likely to use this a lot for the same values of N, we'll memoize the constants we
    use for multiplying.

    """
    N				= len( y )
    #assert N <= 170, f"idft: {N=} too large."  # 512-bit entropy as 3-bit symbols
    e_sk_N			= idft.e_sk.setdefault( N, [] )
    if e_sk_N:
        e_sk_N_i		= iter( e_sk_N )
        result			= [
            sum(
                y_k * e_sk
                for y_k,e_sk in zip( y, e_sk_N_i )
            ) / N+0j
            for n in range( N )
        ]
    else:
        #print( f"idft: Memoizing {N=} multiplication factors" )
        result			= []
        for n in range( N ):
            r			= 0
            s			= 2j * cmath.pi * n / N
            for k in range( N ):
                e_sk		= cmath.exp( s * k )
                e_sk_N.append( e_sk )
                r	       += y[k] * e_sk
            r		       /= N+0j
            result.append( r )
    return result
idft.e_sk			= {}  # noqa: E305


def dft_magnitude( bins: Sequence[complex] ) -> List[float]:
    return [ abs( b ) for b in bins ]


def dft_normalize( bins: Sequence[complex] ) -> List[float]:
    N				= len( bins )
    assert N % 2 == 0, \
        "Only even numbers of DFT bins are supported, not {len(x)}"
    scale			= math.log( N, 2 )
    return [ b / scale for b in bins ]


def dft_to_rms_mags( bins: List[complex] ) -> List[float]:
    """Compute the real-valued magnitudes of the provided complex-valued DFT bins.  Since real samples
    cannot distinguish between a +1Hz and a -1Hz signal, the bins assigned to all intervening
    frequencies between the DC and max bins must be combined.

    Only even numbers of bins are supported.  The resultant number of DFT bins is N/2+1, ordered
    from DC to max.

    And, while we're at it, we'll normalize the scale so that DFTs of one number of bins is
    comparable to another; they are scaled at a rate of sqrt(N).  This makes the RMS energy and
    magnitudes on DFTs of different sizes comparable with eachother.

    """
    N				= len( bins )
    mags			= dft_magnitude( bins )
    norm			= dft_normalize( mags )
    nrms			= rms( norm )
    for i in range( 1, N//2 ):
        norm[i]		       += norm[-i]
        norm[-i]		= 0.0
    return nrms, norm[:N//2+1]


def dft_on_real( bins ):
    nrms, mags			= dft_to_rms_mags( bins )
    return mags


def denoise_mags( mags, threshold, middle=None, stride=8 ):
    """Look for top signals within the given magnitudes.  This involves finding a noise floor within the
    magnitude bins w/ a signal level threshold x above the noise floor.  We'll compute the SNR of
    each signal by iteratively scaling the largest signals down to the average, and then look for
    new signals that have risen above the new (lower) noise-floor avg.  Stops when the remaining
    bins are below the avg * threshold.

    Returns the noise threshold target, and list of the top mags: (target, [ (snr,index), ... ])

    """
    snrs			= set()
    curs			= mags[:]
    cavg			= avg( curs ) if middle is None else middle( curs )
    target			= cavg * threshold
    while peaks := set(
            i
            for i,m in enumerate( curs )
            if i not in snrs and m >= target
    ):
        # There are 1 or more new peaks above the middle curs 'cavg' value.  Note that they are
        # candidate signals in snrs.
        snrs.update( peaks )
        # Replace each signal with the current cavg * threshold 'target'.  This lowers the "avg"
        # noise level by removing the signal portion of the now recognized signal bins; only the
        # remaining non-signal portion of the signal candidate bins and the remaining non-gisnal
        # bins influence the new cavg.
        curs			= [
            target if i in snrs else b
            for i, b in enumerate( mags )
        ]
        cavg			= avg( curs ) if middle is None else middle( curs )
        target			= cavg * threshold
        #print( f"dnoi: {' '.join( f'{c:{stride}.1f}({mags[i] / target if target else 0:{stride-2}.1f})' for i,c in enumerate( curs ))}: {target=:7.1f}" )

    # Return the noise threshold target magnitude deduced, and the index,SNR for each signal bin,
    # sorted by SNR
    return target, sorted(
        ( (i, mags[i] / target if target else 0) for i in snrs ),
        key	= lambda e: e[1],
        reverse	= True,
    )


def signal_recover_real( dfts, scale=None, integer=False, amplify=None ):
    """Recover a real signal from the provided DFT output.  Optionally, scale the DFT before
    recovering the signal.  Optionally, round the values to integer (eg. if the original signal was
    integer).

    """
    #print( "dfts: " + ' '.join( f"{d:{stride*2}.1f}" for d in dfts ))
    assert scale in (None, 0, 1) or scale % 2 == 0, \
        f"Only even multiples of a DFT are valid, not x{scale}"
    while ( scale or 1 ) >= 2:
        scale		      //= 2
        N			= len( dfts )
        # When doubling the length of a DFT, we have to scale each entry by 2x
        dfts			= [d*2 for d in dfts]
        # Insert N entries in the middle
        dfts			= dfts[:N//2] + [0j] * N + dfts[N//2:]
        # Split the high-frequency bin from the old DFT between the 2 new lower frequency bins.  Was
        # in position [N//2+1], or also [-N//2]; now is stil in the position [-N//2], since N is
        # still the old DFT size!  The corresponding location is at dft[N//2]; this will always be a
        # freshly added 0+0j bin.
        dfts[-N//2]	       /= 2
        dfts[N//2]	       += dfts[-N//2]
    #print( "dfts: " + ' '.join( f"{d:{stride*2}.1f}" for d in dfts ))
    sigR			= ifft( dfts )
    #print( "sigR: " + ' '.join( f"{s:{stride*2}.1f}" for s in sigR ))
    sigR			= [ s.real for s in sigR ]
    #print( "sigR: " + ' '.join( f"{s:{stride*2}.1f}" for s in sigR ))
    if amplify:
        sigR			= [s * amplify for s in sigR]
    if integer:
        sigR			= list( map( int, map( round, sigR )))
        #print( "sigR: " + ' '.join( f"{s:{stride*2}}" for s in sigR ))
    return sigR


def signal_draw( s, scale=None, neg=None, pos=None ):
    """Draws a single-line signed waveform if pos/neg is None, or the +'ve/-'ve half
    of a waveform.  Default scale is 1 signed byte.

        default:    "'‾~_.,._~‾'
        pos:                     ,._~‾'"'‾~_.,
        neg:        "'‾~_.,._~‾'"

    """
    if neg is True or pos is False:
        return " \"'‾~_.,"[max( -1, min( 6, int( (-s-1) * 7 // ( scale or 128 )))) + 1]
    elif pos is True or neg is False:
        return " ,._~‾'\""[max( -1, min( 6, int(   s    * 7 // ( scale or 128 )))) + 1]
    else:
        return ",._~‾'\"" [max( -3, min( 3, round( s    * 7  / ( scale or 256 )))) + 3]


if __name__ == "__main__":
    x				= [ 2, 3, 5, 7, 11 ]
    print( "vals:   " + ' '.join( f"{f:11.2f}" for f in x ))
    y				= fft( x )
    print( "DFT:    " + ' '.join( f"{f:11.2f}" for f in y ))
    z				= ifft( y )
    print( "inverse:" + ' '.join( f"{f:11.2f}" for f in z ))
    print( " - real:" + ' '.join( f"{f.real:11.2f}" for f in z ))

    N				= 8
    print( f"Complex signals, 1-4 cycles in {N} samples; energy into successive DFT bins" )
    for rot in (0, 1, 2, 3, -4, -3, -2, -1):  # cycles; and bins in ascending index order
        if rot > N/2:
            print( "Signal change frequency exceeds sample rate and will result in artifacts" )
        sigs                    = [
            # unit-magnitude complex samples, rotated through 2Pi 'rot' times, in N steps
            cmath.rect(
                1, cmath.pi*2*rot/N*i
            )
            for i in range( N )
        ]
        print( f"{rot:2} cycle" + ' '.join( f"{f:11.2f}" for f in sigs ))
        dfts                    = fft( sigs )
        print( "  DFT:  " + ' '.join( f"{f:11.2f}" for f in dfts ))
        print( "   ABS: " + ' '.join( f"{abs(f):11.2f}" for f in dfts ))


# dB <-> ratio calculations.  Default dB definitions are in terms of
# power.  However, most measured values are of fields (eg. voltages).
dB_type_kwds = {
    'power': {              # eg. Antenna dBi gain
        'base':   10.0,
        'factor': 10.0,
    },
    'field': {              # eg. Amplifier voltage gain
        'base':   10.0,
        'factor': 20.0,
    },
    'sensed':  {            # eg. Loudness sensation gain
        'base':    2.0,
        'factor': 10.0,
    },
}


# dB <-> ratio (for power, by default).  If reference is 1 (default),
# then the value is a ratio.  Otherwise, the 'dB' of the 'value'
# vs. the 'reference' value is computed.  This is the definition of
# dB, so deals in terms of Power (Energy) values by default.
def value_from_dB( dB, reference=1, base=10.0, factor=10.0 ):
    return base ** ( dB / factor ) * reference


def dB_from_value( ratio, reference=1, base=10.0, factor=10.0 ):
    """Convert a ratio [0,oo) into a dB range (-oo,oo), where the ratio 1.0 == 0.0dB.  A -'ve or 0
    is invalid, and tends to -oo (negative infinity).

    """
    if reference and reference != 1:
        ratio		       /= reference
    if ratio > 0:
        return factor * math.log( ratio, base )
    return -math.inf


# dB <-> value (for field values, by default).  This is the practical
# dB, so by default deals in measured values, which are usually Field
# strength measurements (eg. Voltage, Sound Pressure, ...)
def from_dB( dB, reference=1, dB_type='field' ):
    return value_from_dB( dB=dB, reference=reference, **dB_type_kwds[dB_type] )


def into_dB( ratio, reference=1, dB_type='field' ):
    return dB_from_value( ratio=ratio, reference=reference, **dB_type_kwds[dB_type] )


class Signal( namedtuple('Signal', [ 'dB', 'stride', 'symbols', 'offset', 'details' ] )):
    """Provide details on the location of symbols that seem to indicate a reduction in entropy.

    dB:		signal to noise ratio; higher indicates worse entropy (more signal or pattern)
    details:	A textual description of the signal
    """
    def __str__( self ):
        return f"{self.dB:7.2f}dB, at {self.offset=:3}: {self.symbols:2} x {self.stride:2} bits/symbol: {self.details}"


def entropy_bin_ints( entropy_bin, offset, symbols, stride, cancel_dc=None ):
    length			= len( entropy_bin )
    bits			= [
        entropy_bin[off:off+stride]
        for off in range( offset, length - stride + 1, stride )
    ][:symbols]
    ints			= [ int( b, 2 )  for b in bits ]
    if cancel_dc:
        ints			= [ r - 2**(stride-1) for r in ints ]
    # print( "bits: " + ' '.join( f"{b:{stride*2}}" for b in bits ))
    # print( "ints: " + ' '.join( f"{r:{stride*2}}" for r in ints ))
    return ints


def entropy_bin_dfts( entropy_bin, offset, symbols, stride, cancel_dc=True ):
    """Compute the frequency bin magnitudes for 'stride'-bits x symbols starting at bit 'offset'.
    If cancel_dc, shifts the numeric zero from bit value '00...' to '10...'; if the values are
    randomly distributed, the DC offset should be cancelled out.

    """
    sigs			= entropy_bin_ints( entropy_bin, offset=offset, symbols=symbols, stride=stride,
                                                    cancel_dc=cancel_dc )
    #print( "sigs: " + ' '.join( f"{s:{stride*2}}" for s in sigs ))
    dfts			= fft( sigs )

    # See if we can observe any signal/noise ratio on any frequency
    #print( "dfts: " + ' '.join( f"{d:{stride*2}.1f}" for d in dfts ))
    return dfts


def int_decode( c, stride=8 ):
    """Output the decimal (if possible) and decoded view of the integer datum 'c'"""
    if stride == 6:  # base-64 URL-safe
        return f"{'ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789-_'[c]:<{stride}}"
    if stride == 5:  # base-32
        return f"{'ABCDEFGHIJKLMNOPQRSTUVWXYZ234567'[c]:<{stride}}"
    if stride == 4:  # hex nibbles
        return f"x{'0123456789ABCDEF'[c]:<{stride-1}}"
    if stride == 3:  # octal digits
        return f"o{'01234567'[c]:<{stride-1}}"
    if 1 <= stride <= 2:  # bits / base-4
        return f"{'0123'[c]:<{stride}}"
    if stride <= 8 and 32 <= c < 127:
        s		= ("'" + chr( c ) + "'")
        return f"{s:<{stride}}"
    hex			= f"{c:0{(stride+3)//4}X}"
    return f"{hex:<{stride}}" if stride <= len( hex ) + 2 else f"0x{hex:<{stride-2}}"


def signal_entropy(
    entropy: bytes,
    stride: int			= 8,		# bits per symbol
    symbols: Optional[int]	= None,		# symbols per DFT; default to ~128 bits and a power of 2, but < length
    overlap: bool		= False,        # sweep across n-1 bits for symbol start
    threshold: Optional[float]	= None,		# Default: must be a known length/stride/symbols
    ignore_dc: bool		= False,        # For eg. character data, we know there'll be a big DC component
    show_details: bool		= True,		# include signal details (incl. expensive ifft)
    middle: Optional[Callable[List[float],float]] = None,  # default to normalized RMS energy, or eg. statistics.median on bin magnitudes
    harmonics_max: int		= 2,		# maximum harmonics to represent
) -> Optional[Signal]:
    """Checks for signals in the supplied entropy.  If n-bit 'stride' or 'base'-n symbols are found
    to exhibit significant signal pattern over the average noise, then the pattern will be reported.

    Returns the Signal w/ dB signal multiple by which the greatest signal exceeded the threshold
    noise floor (defaults to RMS average energy), and its bit stride and offset from start of
    entropy.

    Therefore, if a signal is found w/ power above the threshold (0.0dB), it will have a +'ve dB.
    Acceptable (low signal strength == high entropy) signals below the threshold will have a -'ve
    dB.  Using something like statistics.median is problematic for low-noise signals, because if a
    sufficient number of bins have 0 energy, the noise will be 0, and a "default" SNR (signal/noise
    *ratio*) of 1.0 will be used (0.0 dB), allowing any more legitimate signal to be used instead of
    the unlikely "perfect" noise-free signal (or signals w/ *only* a DC component if we ignore_dc.)

    So, the result can be sorted by greatest power, and if any power is found to be > 0.0dB, the
    entropy can be rejected as having too much "signal" vs "noise" (entropy).

    """
    entropy_hex			= codecs.encode( entropy, 'hex_codec' ).decode( 'ascii' )
    entropy_bin			= ''.join( f"{int(h,16):0>4b}" for h in entropy_hex )
    length			= len( entropy_bin )
    assert not overlap or not ignore_dc, \
        "Cannot specify both overlap and ignore_dc (intended for handling fixed-location symbols)"
    if symbols is None:
        # Default to the largest even-numbered amount of symbols that will fit in the signal, except
        # if 'overlap' is specified, ensure we sweep across all bit-offsets of the symbol once.
        symbols			= length // ( stride * 2 ) * 2
        if overlap and length - symbols * stride < stride - 1:
            symbols	       -= 2
    assert symbols // 2 * 2 == symbols, \
        f"An even number of symbols must be specified, not {symbols=}"
    assert symbols * stride <= length, \
        f"{symbols} x {stride}-bit symbols is beyond signal {length=}"
    if threshold is None:
        threshold		= signal_entropy.signal_limits.get(
            overlap, {} ).get(
                length, {} ).get(
                    stride, {} ).get(
                        symbols )
    assert threshold and ( 0 < threshold ), \
        f"A small +'ve ratio threshold of Signal energy (0,...) is required for {length}-bit entropy w/ {stride}-bit symbols eg. 300%, not {threshold=!r}"
    #print( f"signal_entropy: {length:3}-bit entropy w/ {symbols:3} x {stride:2}-bit symbols ({ignore_dc=:5}): {threshold=:f}" )
    dc				= 0+0j
    strongest			= None
    mags_all			= []
    for symb in range(0, length - symbols * stride + 1, stride ):
        for slip in range( stride if overlap else 1 ):  # noqa: E111
            offset		= symb + slip
            #print( f"=> {symb=:3} + {slip=:3} == {offset}" )
            if length - offset < symbols * stride:
                break
            dfts		= entropy_bin_dfts( entropy_bin, offset, symbols, stride, cancel_dc=not ignore_dc )
            #print( f"dfts: {' '.join( f'{b:{stride*2}.1f}' for b in dfts )}" )
            dc			= dfts[0]
            if ignore_dc:
                dfts[0]		= 0+0j
            nrms, mags		= dft_to_rms_mags( dfts )  # abs energy bins, from DC to max freq
            mags_all.append( mags )
            #print( f"mags: {' '.join( f'{m:{stride*2}.1f}' for m in mags )}: {sum(mags):7.2f} sum, {avg(mags):7.2f} avg, {nrms:7.2f} RMS; dc: {dc:11.1f} == {abs(dc):7.2f} abs" )
            target, snrs	= denoise_mags( mags, threshold )
            snrd		= dict( snrs )  # i: snr
            #print( f"snrs: {' '.join( f'{snrd[i]:{stride*2}.1f}' if i in snrd else (' ' * stride*2) for i in range( len( mags )))}: {target=:7.1f}" )

            # When we have multiple signals, a Signal with several strong signals should have an SNR
            # higher than something with fewer/weaker signals.  However, only the portion *above*
            # the target threshold count.  So, sum the tops, and compute the overall Signal SNR.
            if snrd:
                peak		= target + sum( mags[i] - target for i in snrd )
            else:
                peak		= max( mags )  # No signals; SNR is greatest bin vs. threshold target
            snr			= ( peak / target ) if target else 1.0
            snr_dB		= into_dB( snr )
            #print( f"tops: {offset=:3}, {peak=:7.2f}, {threshold=:7.2f}, {target=:7.2f}, {snr=:7.2f}, {snr_dB=:7.2f}" )
            if strongest and snr_dB <= strongest.dB:
                continue
            if snr_dB < 0 or not show_details:
                strongest	= Signal( dB=snr_dB, stride=stride, symbols=symbols, offset=offset, details='' )
                continue

            # Find the strongest signal frequency bin.  The max frequency (last) bin indicates some
            # pattern sensed in every second symbol (the Nyquist rate, sampled at 2x the max
            # frequency detectable).  The min frequency (first) bin indicates a DC offset (symbol
            # values not centered around zero).  For N symbols, there are N/2+1 bins, [DC], [min],
            # ..., [max]: the intervening N/2 bins [min], ..., [max] contain evenly spaced
            # frequencies; eg. for N==8, N/2+1==5 bins, N/2==4 min-max bins where:
            #
            #     For symbols ==  8 ==>  64 bits 	For symbols == 16 ==> 128 bits
            #     ------------------------------        ------------------------------
            #     [0](DC)  ==> 8*8== 64 bits/beat       [0](DC)  ==>16*8==128 bits/beat
            #     [1](min) ==> 4*8== 32 bits/beat       [1](min) ==> 8*8== 64 bits/beat
            #     [2]          3*8== 24                 [2]          7*8== 56
            #     [3]          2*8== 16                 [3]          6*8== 48
            #     [4](max) ==> 1*8==  8 bits/beat       [4]          5*8== 40
            #                                           [5]          4*8== 32
            #                                           [6]          3*8== 24
            #                                           [7]          2*8== 16
            #                                           [8](max) ==> 1*8==  8 bits/beat
            #

            # Draw the signal area of interest over 'symbols' of the 'stride'-bit symbols beginning at bit 'offset'.
            details		= '\n'
            offpref		= ''
            if offset > 8:
                details	       += f"...x{offset:<3}>{entropy_bin[offset:]}\n"
                offpref		= ' ' * 8
            else:
                details	       += f"{entropy_bin}\n"
                offpref		= ' ' * offset
            details	       += offpref + ''.join(
                '-_'[(( i-offset ) // stride ) % 2] if ( offset <= i < ( offset + symbols * stride )) else ' '
                for i in range( offset, length )
            ) + '\n'

            if stride >= 4:
                details	       += offpref + ''.join(
                    f"{c:<{stride}}"
                    for c in entropy_bin_ints( entropy_bin, offset=offset, symbols=symbols, stride=stride, cancel_dc=not ignore_dc )
                ) + ' decimal\n'
            details	       += offpref + ''.join(
                int_decode( c, stride=stride )
                for c in entropy_bin_ints( entropy_bin, offset=offset, symbols=symbols, stride=stride )
            ) + f" base-{2**(4 if stride >= 7 else stride)}" + ("/ASCII" if stride >= 7 else "") + " decoding\n"

            # Select the one or two highest energy harmonics, and scale the waveform (which is
            # denominated in stride-bit chunks) to the extent needed to cover the binary version of the
            # entropy (then decimate to fit exactly).  So, if the 8-bit signal values are being
            # recovered, we can scale the waveform by 8x.  Also, scale the amplitude by the (removed) DC
            # component; this counteracts the reduction in dynamic range inherent to typical ASCII
            # values (eg. the digits 0-9 are only 8% of the full ASCII range).
            dfts_rec		= [0+0j] * len( dfts)
            harmonic		= []
            for max_i in snrd:
                if harmonics_max and harmonic and len( harmonic ) >= harmonics_max:
                    break
                # The 0'th bin is DC, 1st is the base frequency, and then for eg. 16 symbols, we get
                # DC + 8 bins, where each bin represents a harmonic frequency component repeating
                # every 16/1==16, 16/2==8, 16/3==5+1/3, 16/4==4, 16/5==3+1/5, ..., 16/8==2 (the max
                # Nyquist frequency) symbols.
                harmonic.append( max_i )
                dfts_rec[max_i]	= dfts[max_i]
                if 0 < max_i < len(mags)-1:
                    dfts_rec[-max_i] = dfts[-max_i]		# ... and its symmetrical bin, if not max freq.
            harmonic_freq	= [
                mixed_fraction( len( dfts ), h )
                for h in harmonic
                if h > 0
            ]

            dc_amplify		= None
            if ignore_dc:
                # scale the signal by the ratio of the removed DC to the largest other signal.  Since we know that
                # the other signals were DC "higher" in the original signal, amplifying the signal by this ratio
                # shouldn't exceed the maximum dynamic range of the eg. integer signal values.
                peak		= max( abs( b ) for b in dfts )
                if peak:
                    dc_amplify	= 1 + abs( dc ) / peak

            # Compute the resolution for the inverse DFT required to properly cover the binary
            # entropy.  However, limit size of the resultant DFT, as idft is O(N^2).  For a 512-bit
            # entropy split into 3-bit symbols, results in a N=170 DFT.  That's the 28,900
            # multiplications to produce an inverse-DFT.  That's the absolute largest we want to
            # deal with, and we don't want to produce anything bigger than an N=128 DFT, because we
            # don't need greater resolution than that to produce a smooth function graph in text.
            scale_signal	= 1
            scale_stride	= stride
            while scale_stride > 2 and len( dfts_rec ) * scale_signal * 2 <= 128:
                scale_stride   /= 2
                scale_signal   *= 2
            sigs		= signal_recover_real( dfts_rec, scale=scale_signal, integer=True, amplify=dc_amplify )

            pos			= ''
            neg			= ''
            o			= 0
            for i in range( symbols * stride ):
                o			= int( i / scale_stride )
                #print( f" - {i=:3} --> {o=:3} ==> {sigs[o]:7}" )
                pos	       += signal_draw( sigs[o], pos=True )
                neg	       += signal_draw( sigs[o], neg=True )
            harmonic_dBs	= [ f"{ordinal(h) if h else 'DC'} {into_dB(mags[h]/target) if target else 0.0:.1f}dB" for h in harmonic ]
            details	       += f"{offpref}{pos} {len(harmonic)} harmonics: {commas( harmonic_dBs, final='and' )}\n"
            details	       += f"{offpref}{neg}  - "
            if 0 in harmonic:
                details	       += "DC offset"
            if 0 in harmonic and harmonic_freq:
                details	       += " and "
            if harmonic_freq:
                details	       += f"every {commas( harmonic_freq, final='and' )} symbols"
            details	       += "\n"
            strongest		= Signal( dB=snr_dB, stride=stride, symbols=symbols, offset=offset, details=details )
    #mags_avgs			= [sum(col)/len(mags_all) for col in zip(*mags_all)]
    #print( f"avgs: {' '.join( f'{m:{stride*2}.1f}' for m in mags_avgs )}: {sum(mags_avgs):7.2f}" )
    return strongest

signal_entropy.signal_limits	= {  # noqa: E305
    False: {
        128: {
            3: {
                40: 3.7959290562900327,
                42: 3.647561915551939
            },
            4: {
                30: 3.8431997942042178,
                32: 3.8749244192110694
            },
            5: {
                22: 4.01132938746108,
                24: 3.855818706529049
            },
            6: {
                18: 3.9821710568269393,
                20: 4.10175893691386
            },
            7: {
                16: 3.9998153373627243,
                18: 3.86288147813215
            },
            8: {
                14: 4.035389231470121,
                16: 3.85295803340856
            }
        },
        160: {
            3: {
                50: 3.768307059766632,
                52: 3.657642158466681
            },
            4: {
                38: 3.927175392223398,
                40: 3.644644794089787
            },
            5: {
                30: 4.004859915596088,
                32: 3.723428336867249
            },
            6: {
                24: 3.921311891749215,
                26: 3.7484305351772997
            },
            7: {
                20: 4.053741462343345,
                22: 3.7022202971931337
            },
            8: {
                18: 3.902181587171419,
                20: 3.682978660449477
            }
        },
        192: {
            3: {
                62: 3.78844766566635,
                64: 3.6289494114027234
            },
            4: {
                46: 3.9167065596553523,
                48: 3.6393513794634913
            },
            5: {
                36: 3.8915275898547788,
                38: 3.7473928123694793
            },
            6: {
                30: 3.9132919286186874,
                32: 3.904575332226886
            },
            7: {
                24: 4.18637617875428,
                26: 3.8093336871612897
            },
            8: {
                22: 4.046780450578674,
                24: 3.766307067356347
            }
        },
        224: {
            3: {
                72: 3.6989229751871116,
                74: 3.858442530730556
            },
            4: {
                54: 3.868023231802504,
                56: 3.821289665038688
            },
            5: {
                42: 3.9244827423261506,
                44: 3.7255928738074235
            },
            6: {
                34: 3.9796568789267033,
                36: 3.8424176850637064
            },
            7: {
                30: 3.884348278904745,
                32: 3.7274499501787015
            },
            8: {
                26: 3.9286076413465723,
                28: 3.8212509511408332
            }
        },
        256: {
            3: {
                82: 3.874105542934525,
                84: 3.8031640175138075
            },
            4: {
                62: 3.8535667393555895,
                64: 3.6702455043197517
            },
            5: {
                48: 3.966192695309544,
                50: 3.810215205319617
            },
            6: {
                40: 3.780615203182523,
                42: 3.7166374907758537
            },
            7: {
                34: 3.7632487952840674,
                36: 3.820508681003472
            },
            8: {
                30: 3.851962569866387,
                32: 3.79387401596316
            }
        },
        512: {
            3: {
                168: 3.8510228062858536,
                170: 3.6930501052854368
            },
            4: {
                126: 3.8739687818152095,
                128: 3.8390294681049797
            },
            5: {
                100: 3.8944436960232816,
                102: 3.8213977750827777
            },
            6: {
                82: 3.8651467331943445,
                84: 4.00628842914143
            },
            7: {
                70: 3.8977031690387087,
                72: 3.7538315552341635
            },
            8: {
                62: 3.956950350574148,
                64: 3.7809092998228437
            }
        }
    },
    True: {
        128: {
            3: {
                40: 4.015412815647333,
                42: 4.049142728796211
            },
            4: {
                30: 4.084354511564386,
                32: 3.651281177941025
            },
            5: {
                22: 4.190848444589507,
                24: 4.358983589533379
            },
            6: {
                18: 4.433227305912341,
                20: 4.300266280446989
            },
            7: {
                16: 4.63348257265716,
                18: 4.251514929567565
            },
            8: {
                14: 4.539487223257121,
                16: 3.8298375269653464
            }
        },
        160: {
            3: {
                50: 4.102991330499331,
                52: 3.9201904890263806
            },
            4: {
                38: 3.9660043286866316,
                40: 3.8885457864057296
            },
            5: {
                30: 4.1534554695112424,
                32: 3.875257196943185
            },
            6: {
                24: 4.314705804221271,
                26: 4.276913496937379
            },
            7: {
                20: 4.484017655351361,
                22: 4.1891981747029154
            },
            8: {
                18: 4.403790052686062,
                20: 3.8171358828247257
            }
        },
        192: {
            3: {
                62: 3.946957410520236,
                64: 3.707417210141612
            },
            4: {
                46: 4.243992805910835,
                48: 3.5294430695860983
            },
            5: {
                36: 4.307388423581146,
                38: 3.8978305998977363
            },
            6: {
                30: 4.296222571278912,
                32: 3.666087989048472
            },
            7: {
                24: 4.483230938800909,
                26: 4.2688078347472285
            },
            8: {
                22: 4.3975014624329,
                24: 3.9299336762739103
            }
        },
        224: {
            3: {
                72: 4.020867167290543,
                74: 4.115472579780688
            },
            4: {
                54: 4.155505881349097,
                56: 3.611932501064818
            },
            5: {
                42: 4.296142369450004,
                44: 4.075549039418728
            },
            6: {
                34: 4.339759784005132,
                36: 4.199962133110188
            },
            7: {
                30: 4.323936203605842,
                32: 4.029221841050185
            },
            8: {
                26: 4.470823290289282,
                28: 3.8214418599675444
            }
        },
        256: {
            3: {
                82: 3.9159493966982546,
                84: 3.9052586437888848
            },
            4: {
                62: 4.095016786713433,
                64: 3.6867364579681587
            },
            5: {
                48: 4.260740010786478,
                50: 4.152970392617726
            },
            6: {
                40: 4.298675707470258,
                42: 3.9884678302777594
            },
            7: {
                34: 4.218469498120002,
                36: 4.09636224540939
            },
            8: {
                30: 4.299777044591192,
                32: 3.633717478015482
            }
        },
        512: {
            3: {
                168: 4.016185108804742,
                170: 4.028203232067135
            },
            4: {
                126: 4.039199984086117,
                128: 3.694796108995797
            },
            5: {
                100: 4.012546598293882,
                102: 3.8791966508371343
            },
            6: {
                82: 4.071645334320387,
                84: 4.31042853254462
            },
            7: {
                70: 4.1408664186751505,
                72: 4.060968141932326
            },
            8: {
                62: 4.290524429312738,
                64: 3.8260046870654842
            }
        }
    }
}


def shannon_entropy(
    entropy: bytes,
    stride: int			= 8,		# bits per symbol
    overlap: bool		= True,
    threshold: float		= None,         # Allow up to a certain % bits-per-symbol entropy deficit
    show_details: bool		= True,		# include signal details (incl. expensive ifft)
    snr_min: float		= 1/100,        # Minimum snr .01 == -40dB (eg. if all symbols unique)
    N: Optional[int]		= None,		# If a max. number of unique symbols is known eg. dice
) -> Optional[Signal]:
    """Estimates the "Symbolised Shannon Entropy" for stride-bit chunks of the provided entropy; by
    default, since we don't know the bit offset of any pattern, we'll scan at each possible bit
    offset (overlap = True).

    We'll compute the estimated predictability of the entropy, in the range [0,1], where 0 is
    unpredictable (high entropy), and 1 is totally predictable (no entropy), but as a dB range,
    where -'ve dB is below the threshold and acceptable, and +'ve is above the predictability
    threshold, and is unacceptable.

    Typically, for large entropy and/or small 'stride' values, the number of unique values N found
    should tend toward 2^stride.  However, for small entropy (eg. insufficient data) or small known
    N (eg. die rolls), we expect to see less unique values than would be implied by the stride.
    Therefore, we reduce the N in those cases.

    The "strongest" (least entropy, most predictable, most indicative that the entropy is
    unacceptable) signal will be returned.

    See: https://www.sciencedirect.com/topics/engineering/shannon-entropy

    We provide some known defaults for evaluating data of various sizes, which are tested to produce
    only about a 1% rejection rate for good entropy.

    """
    entropy_hex			= codecs.encode( entropy, 'hex_codec' ).decode( 'ascii' )
    entropy_bin			= ''.join( f"{int(h,16):0>4b}" for h in entropy_hex )
    length			= len( entropy_bin )

    # Find all the unique n-bit symbols in the entropy at the desired offset(s)
    strongest			= None
    for offset in range( stride ) if overlap else (0,):
        # Calculate the frequency of each unique symbol
        frequency		= defaultdict( int )
        # How many full n-bit symbols fit in the entropy at the current symbol-start offset?
        symbols			= ( length - offset ) // stride
        for s in range( symbols ):
            i			= offset + s * stride
            symbol		= entropy_bin[i:i+stride]
            frequency[symbol]  += 1
        assert sum( frequency.values() ) == symbols, \
            f"Expected {symbols=} sum of probabilities events, found {frequency!r}"
        assert all( len( s ) == stride for s in frequency ), \
            f"Expected all {stride}-bit symbols, found {frequency!r}"

        bitspersymbol			= -sum(  # sum may range: (~-0.0,...)
            probability/symbols * math.log( probability/symbols, 2 )
            for probability in frequency.values()
        )
        # For small numbers of symbols, we cannot achieve a full N unique samples.  Therefore, the
        # number of "bits of entropy" per symbol will be low -- even if all of the samples obtained
        # are unique.  Therefore, we return a value in the range (0=bad,1=good) scaled by the bits
        # required to encode the maximum number of unique symbols possible (due to the symbol or
        # entropy size), not in the range [0,2^stride].
        shannon			= 0.0
        N_min			= min( N or symbols, symbols, 2**stride )
        assert len( frequency ) <= N_min, \
            f"Observed {len(frequency)} unique symbols; more than expected by min({N=}, {symbols=} or {2**stride=})"
        if N_min > 1:
            shannon		= bitspersymbol / math.log( N_min, 2 )

        # So now, shannon is 1.0 for a "good" perfectly unpredictable entropy (all symbols
        # different, full bits-per-symbol required to encode), and 0 for "bad" perfectly predictable
        # entropy (all symbols identical, zero bits-per-symbol to encode).  This is the inverse of
        # what we want: the more predictable something is, the stronger the "signal" is vs. the
        # noise (entropy).  A 10% threshold means "I want at most 10% predictability; I want 90%
        # noise".  This
        predictability		= 1 - shannon			# 0 ==> good entropy, 1 ==> no entropy

        # We want a -'ve dB if below (low predictability, good entropy, acceptable), +'ve if
        # at/above threshold (high predictability, too much signal, reject).  1.0x == 0.0dB.
        # However, we'll typically often see good, high-entropy data which turns out to have a
        # shannon == 1.0, implying a predictability == 0, yielding an snr ratio of 0, which is
        # invalid (tends to -oo).  We want to avoid this.  We know that the range is predictability
        # range is (0,1), and we'll typically allow a threshold of 10%, so 0 is not -oo!  We really
        # want to map (0...threshold,...1) onto a ratio range like .01 (-40dB) to 100 (+40dB), with
        # 1.0 == 0dB right at threshold.  Map the range (0,threshold] to (.01,1] (mapped to -'ve
        # dB), and let the +'ve dB (threshold,1) range fall where it may.  For example, if threshold
        # is .1 and predictability is also .1, we want to map that to exactly 1.0:
        #
        #             predictability
        #             v
        #     1/100 + 0    / ( .1 / ( 1 - 1/100 )) == 0.0100 == -10.0   dB  # -'ve below threshold
        #     1/100 + .043 / ( .1 / ( 1 - 1/100 )) == 0.4357 ==  -7.216 dB  #
        #     1/100 + .1   / ( .1 / ( 1 - 1/100 )) == 1.0000 ==   0.0   dB  # exactly at threshold
        #     1/100 + 1    / ( .1 / ( 1 - 1/100 )) == 9.9100 ==  19.20  dB  # +'ve above threshold
        #
        thresh			= shannon_entropy.shannon_limits.get(
            overlap, {} ).get(
                length, {} ).get(
                    stride ) if threshold is None else threshold
        assert thresh and 0 < thresh <= 2, \
            f"A small +'ve ratio threshold of Shannon entropy deficit ~ (0, 1] is required for {length}-bit entropy w/ {symbols} x {stride}-bit symbols eg. 10%, not {thresh!r}"

        snr			= snr_min + predictability / ( thresh / ( 1 - snr_min ))
        snr_dB			= into_dB( snr )
        weaker			= strongest and snr_dB < strongest.dB
        ( log.debug if snr_dB < 0 else log.info )(
            f"Found {len(frequency):3} unique (of {N_min:3} possible) in {symbols:3}"
            f"x {stride:2}-bit symbols at offset {offset:2} in {length:4}-bit entropy:"
            f" Shannon Entropy {bitspersymbol:7.3f} b/s, P({shannon:7.3f}) unpredictable; {predictability=:7.3f}"
            f" vs. threshold={thresh:7.3f} == {snr=:7.3f} {snr_dB:7.3f}dB: {entropy_hex}"
        )
        if weaker:
            continue
        longest			= max(len(s) for s in frequency)
        details			= ''
        if snr_dB >= 0 and show_details:
            details	       += f"{len(frequency):2} unique"
            interesting		= sorted(
                frequency.items(), reverse=True, key=lambda kv: kv[1]
            )
            if stride < 7:
                details	       += f" (base-{2**stride})"
            details	       += ": " + commas(
                f"{v:>{longest}} = {int_decode( int( v, 2 ), stride=stride ).strip()}: {c:2}"
                for v,c in ( interesting[:4] if len( interesting ) > 5 else interesting )
            )
            if len( interesting ) > 5:
                details	       += f", ...x{len( interesting ) - 6}, " + commas(
                    f"{v:>{longest}} = {int_decode( int( v, 2 ), stride=stride ).strip()}: {c:2}"
                    for v,c in interesting[-2:]
                )
            details	       += "\n"
            # Show the frequency of the symbols
            offpref		= ''
            if offset > 8:
                details	       += f"...x{offset:<3}>{entropy_bin[offset:]}\n"
                offpref		= ' ' * 8
            else:
                details	       += f"{entropy_bin}\n"
                offpref		= ' ' * offset

            most		= interesting[0][1]
            least		= min( interesting[-1][1], most - 1 )
            scale		= 255 // ( most - least )
            wav			= ''
            for s in range( symbols ):
                i		= offset + s * stride
                symbol		= entropy_bin[i:i+stride]
                signal		= ( frequency[symbol] - least ) * scale - 128
                wav	       += stride * signal_draw( signal )
            details	       += f"{offpref}{wav}\n"

        strongest		= Signal(
            dB		= snr_dB,
            stride	= stride,
            symbols	= symbols,
            offset	= offset,
            details	= details,
        )
    return strongest

shannon_entropy.shannon_limits	= {  # noqa: E305
    False: {
        128: {
            3: 0.13774543485038598,
            4: 0.21352083735742364,
            5: 0.25857335693364575,
            6: 0.18024155906953593,
            7: 0.13393449488438314,
            8: 0.10565473026867159
        },
        160: {
            3: 0.10849932592752157,
            4: 0.1732988370721074,
            5: 0.2539763216172974,
            6: 0.17643151270139718,
            7: 0.13155406158983765,
            8: 0.09785468929753503
        },
        192: {
            3: 0.09151054889558748,
            4: 0.14093100938467484,
            5: 0.2264278561929118,
            6: 0.1760839014542283,
            7: 0.1271444744051939,
            8: 0.08862800898521093
        },
        224: {
            3: 0.07916495360459236,
            4: 0.12252481621910122,
            5: 0.20074807382273818,
            6: 0.18204659485920255,
            7: 0.12255892501570535,
            8: 0.08642238826221553
        },
        256: {
            3: 0.06465367140204993,
            4: 0.11144594962520256,
            5: 0.178731925009808,
            6: 0.1819179694876547,
            7: 0.12157462190503683,
            8: 0.0860841002032892
        },
        512: {
            3: 0.03404772973774162,
            4: 0.05292514037032153,
            5: 0.09170667883453174,
            6: 0.15075753744881315,
            7: 0.1264033310600746,
            8: 0.08353573387816829
        }
    },
    True: {
        128: {
            3: 0.15298580418729513,
            4: 0.2451904173880027,
            5: 0.2830395399608447,
            6: 0.2028328095184408,
            7: 0.16471175054429602,
            8: 0.13936432438214608
        },
        160: {
            3: 0.12038637754457157,
            4: 0.19979446717148586,
            5: 0.28247744681759446,
            6: 0.19821636321414943,
            7: 0.15105344745335572,
            8: 0.1174014993567044
        },
        192: {
            3: 0.09947716314352242,
            4: 0.16262594321420024,
            5: 0.24061446698597114,
            6: 0.1988991682768491,
            7: 0.14865256303627913,
            8: 0.11409493984939509
        },
        224: {
            3: 0.09107077508129978,
            4: 0.13949972486012852,
            5: 0.21449493978928671,
            6: 0.18980489337861572,
            7: 0.14016478187097317,
            8: 0.10760893832461406
        },
        256: {
            3: 0.07641224240206347,
            4: 0.11638307614673703,
            5: 0.19375895117831868,
            6: 0.19231952709666766,
            7: 0.13550085464769412,
            8: 0.09588448833768029
        },
        512: {
            3: 0.038422756462380414,
            4: 0.057658537793532146,
            5: 0.09918546794102677,
            6: 0.16114705850045197,
            7: 0.1402080308731611,
            8: 0.09257360822755922
        }
    }
}


def scan_entropy(
    entropy: bytes,
    strides: Optional[Union[int,Tuple[int,int]]] = None,  # If only a specific stride/s makes sense, eg. for ASCII symbols
    overlap: bool		= True,
    ignore_dc: bool		= False,
    show_details: bool		= True,
    N: Optional[int]		= None,			# shannon_entropy may specify limited unique symbols
    signal_threshold: Optional[float]	= None,
    shannon_threshold: Optional[float]	= None,
) -> Tuple[List[Signal],List[Signal]]:
    """Defaults to as many symbols as we can manage, given 'overlap' (which ensures we scan scan at
    least a full stride of bit offsets).

    There is usually little benefit to performing both with/without overlap; we still analyze the
    0-offset symbols, but perhaps with a couple fewer total symbols, and with a slightly higher
    threshold.

    """
    if strides is None:
        strides			= (3, 9)
    else:
        try:
            _,_			= strides
        except TypeError:
            strides		= (int(strides), int(strides)+1)

    signals			= sorted(
        (
            s
            for s in (
                signal_entropy( entropy, stride=stride, overlap=overlap, ignore_dc=ignore_dc,
                                show_details=show_details, threshold=signal_threshold )
                for stride in range( *strides )
            )
            if s.dB >= 0
        ), reverse=True )
    shannons			= sorted(
        (
            s
            for s in (
                shannon_entropy( entropy, stride=stride, overlap=overlap, N=N,
                                 show_details=show_details, threshold=shannon_threshold )
                for stride in range( *strides )
            )
            if s.dB >= 0
        ), reverse=True )
    return signals, shannons


def display_entropy( signals, shannons, what=None ):
    result			= None
    if signals or shannons:
        report			= ''
        dBs			= sorted( list( s.dB for s in signals ) + list( s.dB for s in shannons) )
        dBr			= dBs[:1] + dBs[max( 1, len(dBs)-1):]
        report		       += f"Entropy analysis {('of ' + what) if what else ''}: {len(signals)+len(shannons)}x"
        report		       += f" {'-'.join( f'{dB:.1f}' for dB in dBr )}dB non-random patterns in"
        report		       += f" {commas( sorted( set( s.stride for s in signals )), final='and' )}-bit symbols\n\n"
        for s,summary in sorted(
                [
                    (s, f"{s.dB:5.1f}dB Signal harmonic feature at offset {s.offset} in {s.symbols}x {s.stride}-bit symbols")
                    for s in signals
                ] + [
                    (s, f"{s.dB:5.1f}dB Shannon entropy reduced at offset {s.offset} in {s.symbols}x {s.stride}-bit symbols")
                    for s in shannons
                ], reverse=True ):
            report	       += f"-{summary}{': ' if s.details else ''}\n"
            if s.details:
                report	       += f"{s.details}\n"
        result			= report
    return result


def analyze_entropy(
    entropy: bytes,
    strides: Optional[Union[int,Tuple[int,int]]] = None,  # If only a specific stride/s makes sense, eg. for ASCII symbols
    overlap: bool		= True,
    ignore_dc: bool		= False,
    what: Optional[str]		= None,
    show_details: bool		= True,
    N: Optional[int]		= None,			# shannon_entropy may specify limited unique symbols
    signal_threshold: Optional[float]	= None,
    shannon_threshold: Optional[float]	= None,
) -> Optional[str]:
    """Analyzes the provided entropy.  If patterns are found, reports the findings; the peak Signal
    and the aggregate report: (Signal, "...").

    Seek strong Signals or weak Shannon Entropy, across a number of different interpretations (bit
    strides and overlaps) of the entropy data.  We do not know where poor entropy may hide...

    Since the probability of a signal being found is:

        P( A or B or ... or Z ) = P(A) + P(B) + ... + P(Z) - P(A and B and ... Z )
      = P( A or B or ... or Z ) = P(A) + P(B) + ... + P(Z) - ( P(A) * P(B) * ... P(Z) )

    since we set the {signal,shannon}_entropy threshold values at ~ <1%, we know the P(A and B...)
    term is very small; the probability we'll find an entropy failure it basically is the sum of the
    individual probabilities of each test failing.

    So, if we try 6 of each analysis, that's 12 * 1% =~= 12%.  If we want a total failure of about
    1%, so we must target a ~0.1% failure on good entropy for each test.

    However, it seems likely that if a signal appears in one bit-stride, it may appear in others, so
    perhaps the analysis on different strides may be related.  So, it might be practical to target
    0.25% failure on each individual test.

    """
    return display_entropy(
        *scan_entropy(
            entropy, strides, overlap, ignore_dc=ignore_dc, show_details=show_details, N=N,
            signal_threshold=signal_threshold, shannon_threshold=shannon_threshold ),
        what=what
    )



import pytest
import http.client
import json
import os


# We don't want to run this test unless we provide our bitquery API
# key in X-API-KEY
@pytest.mark.skipif( not os.getenv( "BITQUERY_API_KEY" ), reason="Missing BITQUERY_API_KEY environment variable" )
def test_bitquery_smoke():

    queries			= [
        {
            "query": """\
query ($network: EthereumNetwork!, $addresses: [String!]) {
  ethereum(network: $network) {
    address(address: {in: $addresses}) {
      address
      annotation
      balances {
        value
        currency {
          address
          symbol
          tokenType
        }
      }
    }
  }
}
""",
            "variables": {
                "network": "ethereum",
                "addresses": [
                    "0x22615C3A31d8f9d47bdB84502780A8D2C136fCF5"
                ]
            }
        },
        {
            "query": """\
query ($network: BitcoinNetwork!, $addresses: [String!]) {
  bitcoin(network: $network) {
    inbound: coinpath(receiver: {in: $addresses}) {
      receiver {
        address
      }
      amount
    }
  }
}
""",
            "variables": {
                "network": "bitcoin",
                "addresses": [
                    "bc1qcj9ujyvrf94wu0902g2lnklzlyn5j5nrr44hwp",
                    "18cBEMRxXHqzWWCxZNtU91F5sbUNKhL5PX",
                    "bc1qygm3dlynmjxuflghr0hmq6r7wmff2jd5gtgz0q"
                ]
            }
        }
    ]

    headers = {
        'Content-Type': 'application/json',
        'X-API-KEY': os.getenv( "BITQUERY_API_KEY" ),
    }

    conn = http.client.HTTPSConnection("graphql.bitquery.io")

    for query in queries:
        payload = json.dumps( query )
        print( payload )

        conn.request("POST", "/", payload, headers)
        rx = conn.getresponse()
        rxstr = rx.read().decode("UTF-8")
        print(rxstr)
        response = json.loads( rxstr )
        print( json.dumps( response, indent=4 ))




#
# Python-slip39 -- Ethereum SLIP-39 Account Generation and Recovery
#
# Copyright (c) 2022, Dominion Research & Development Corp.
#
# Python-slip39 is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.  It is also available under alternative (eg. Commercial) licenses, at
# your option.  See the LICENSE file at the top of the source tree.
#
# Python-slip39 is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
from __future__		import annotations

import colorsys
import getpass
import logging
import math
import sys
import traceback

from typing		import Union
from fractions		import Fraction
from functools		import wraps


__author__                      = "Perry Kundert"
__email__                       = "perry@dominionrnd.com"
__copyright__                   = "Copyright (c) 2022 Dominion Research & Development Corp."
__license__                     = "Dual License: GPLv3 (or later) and Commercial (see LICENSE)"

log				= logging.getLogger( "util" )

# util.timer
#
# Select platform appropriate timer function
#
if sys.platform == 'win32' and sys.version_info[0:2] < (3,8):
    # On Windows (before Python 3.8), the best timer is time.clock
    from time		import clock	as timer
else:
    # On most other platforms the best timer is time.time
    from time		import time	as timer


#
# @util.timing
#
def timing( fun, instrument=False, level=None ):
    """A timing decorator which optionally instruments the function return value to return
    (duration,value), or logs the call duration (at logging.INFO by default, if not instrumenting).

    """
    @wraps( fun )
    def wrap( *args, **kwds ):
        beg			= timer()
        result			= fun( *args, **kwds )
        end			= timer()
        dur			= end - beg
        if level or not instrument:
            logging.log( level or logging.INFO, f"{fun!r} took {dur:.3f}s" )
        if instrument:
            return (dur,result)
        return result
    return wrap


log_cfg				= {
    "level":	logging.WARNING,
    "datefmt":	'%Y-%m-%d %H:%M:%S',
    #"format":	'%(asctime)s.%(msecs).03d %(threadName)10.10s %(name)-16.16s %(levelname)-8.8s %(funcName)-10.10s %(message)s',
    "format":	'%(asctime)s %(name)-16.16s %(message)s',
}

log_levelmap 			= {
    -2: logging.FATAL,
    -1: logging.ERROR,
    0: logging.WARNING,
    1: logging.INFO,
    2: logging.DEBUG,
}


def log_level( adjust ):
    """Return a logging level corresponding to the +'ve/-'ve adjustment"""
    return log_levelmap[
        max(
            min(
                adjust,
                max( log_levelmap.keys() )
            ),
            min( log_levelmap.keys() )
        )
    ]


def log_apis( func ):
    """Decorator for logging function args, kwds, and results"""
    @wraps( func )
    def wrapper( *args, **kwds ):
        try:
            result = func(*args, **kwds)
        except Exception as exc:
            log.warning( f"{func.__name__}( {args!r} {kwds!r} ): {exc}" )
        else:
            log.info( f"{func.__name__}( {args!r} {kwds!r} ) == {result}" )
        return result
    return wrapper


#
# util.is_...		-- Test for various object capabilities
#
def is_mapping( thing ):
    """See if the thing implements the Mapping protocol."""
    return hasattr( thing, 'keys' ) and hasattr( thing, '__getitem__' )


def is_listlike( thing ):
    """Something like a list or tuple; indexable, but not a string or a class (some may have
    __getitem__, eg. cpppo.state, based on a dict).

    """
    return not isinstance( thing, (str,bytes,type) ) and hasattr( thing, '__getitem__' )


#
# @util.memoize		-- Cache function results data based on positional args, and maxage/size
#
def memoize( maxsize=None, maxage=None, log_at=None ):
    """A very simple memoization wrapper based on (immutable) args only, for simplicity.  Any
    keyword arguments must be immaterial to the successful outcome, eg. timeout, selection of
    providers, etc..

    Only successful (non-Exception) outcomes are cached!

    Keeps track of the age (in seconds) and usage (count) of each entry, updating them on each call.
    When an entry exceeds maxage, it is purged.  If the memo dict exceeds maxsize entries, 10% are
    purged.

    Optionally logs when we memoize something, at level log_at.
    """
    def decorator( func ):
        @wraps( func )
        def wrapper( *args, **kwds ):
            now			= timer()
            # A 0 hits count is our sentinel indicating args not memo-ized
            last,hits		= wrapper._stat.get( args, (now,0) )
            if not hits or ( maxage and ( now - last > maxage )):
                entry = wrapper._memo[args] = func( *args, **kwds )
                if log_at and log.isEnabledFor( log_at ):
                    if hits:
                        log.log( log_at, "{} Refreshed {!r} == {!r}".format( wrapper.__name__, args, entry ))
                    else:
                        log.log( log_at, "{} Memoizing {!r} == {!r}".format( wrapper.__name__, args, entry ))
            else:
                entry		= wrapper._memo[args]
                #log.detail( "{} Remembers {!r} == {!r}".format( wrapper.__name__, args, entry ))
            hits	       += 1
            wrapper._stat[args] = (now,hits)

            if maxsize and len( wrapper._memo ) > maxsize:
                # Prune size, by ranking each entry by hits/age.  Something w/:
                #
                #   2 hits 10 seconds old > 1 hits 6 seconds old > 3 hits 20 seconds old
                #
                # Sort w/ the highest rated keys first, so we can just eject all those after 9/10ths of
                # maxsize.
                rating		= sorted(
                    (
                        (hits / ( now - last + 1 ), key)		# Avoids hits/0
                        for key,(last,hits) in wrapper._stat.items()
                    ),
                    reverse	= True,
                )
                for rtg,key in rating[maxsize * 9 // 10:]:
                    # log.detail( "{} Ejecting  {!r} == {!r} w/ rating {:7.2f}, stats: {}".format(
                    #     wrapper.__name__, key, wrapper._memo[key], rtg, wrapper._stat[key] ))
                    del wrapper._stat[key]
                    del wrapper._memo[key]
            return entry

        def stats( predicate=None, now=None ):
            """Return the number of memoized data, their average age, and average hits per memoized entry."""
            if now is None:
                now		= timer()
            cnt,age,avg		= 0,0,0
            for key,(last,hits) in wrapper._stat.items():
                if not predicate or predicate( *key ):
                    cnt	       += 1
                    age	       += now-last
                    avg	       += hits
            return cnt,age/(cnt or 1),avg/(cnt or 1)

        wrapper.stats		= stats

        def reset():
            """Flush all memoized data."""
            wrapper._memo	= dict()		# { args: entry, ... }
            wrapper._stat	= dict()		# { args: (<timestamp>, <count>), ... }

        wrapper.reset		= reset

        wrapper.reset()

        return wrapper

    return decorator


#
# @util.retry		-- Retry w/ exponential back-off, 'til truthy result returned
#
def retry( tries, delay=3, backoff=1.5, default_cls=None, log_at=None, exc_at=logging.WARNING ):
    """Retries a function or method until it returns a truthy value.  If default_cls is None, will
    recycle (keep returning) any prior non-falsey value 'til successful again.  Otherwise, will
    return a default_cls instance (or just its falsey value, eg. False) on failure.

    The 'delay' sets the initial delay in seconds, and 'backoff' (defaults to 1.5x; a 50% increase
    in delay each retry) sets the factor by which the delay should lengthen after each
    failure/falsey value.  The 'backoff' multiple must be greater than 1, or else it isn't really a
    backoff.  The 'tries' must be at least 0 to have any exponential effect, and delay greater than
    0; once 'tries' expires, the backoff will remain constant.

    Once truthy, the values reset 'til the next failure.

    """
    if delay <= 0:
        raise ValueError("delay must be greater than 0")
    if backoff <= 1:
        raise ValueError("backoff must be greater than 1")
    if tries < 0:
        raise ValueError("tries must be 0 or greater")

    def decorator( func ):
        @wraps( func )
        def wrapper( *args, **kwds ):
            now			= timer()
            if wrapper.lst is None:
                wrapper.lst	= now
            if wrapper.ok or now >= wrapper.lst + delay * backoff ** min( wrapper.cnt, tries ):
                # Was truthy last time, or it is time to to call again.
                rv		= None
                try:
                    wrapper.lst	= now
                    rv		= func( *args, **kwds )
                except Exception as exc:
                    if exc_at:
                        log.log( exc_at,  f"{wrapper.__name__}: Exception after {wrapper.cnt} {'successes' if wrapper.ok else 'failures'}: {exc}" )
                    if log.isEnabledFor( logging.DEBUG ):
                        logging.debug( "%s", ''.join( traceback.format_exc() ))
                else:
                    if rv:
                        wrapper.rv	= rv
                        return rv
                finally:
                    ok		= bool( rv )
                    if ok != wrapper.ok:
                        # Count resets on rising/falling edge (success <-> failure change)
                        if log_at and log.isEnabledFor( log_at ):
                            log.log( log_at, f"{wrapper.__name__}: Became {'Truthy' if rv else 'Falsey'} after {wrapper.cnt} {'successes' if wrapper.ok else 'failures'}" )
                        wrapper.cnt	= 0		# .cnt resets only on rising/falling edge
                        wrapper.ok	= ok
                    wrapper.cnt += 1			# And *always* counts successes/failures (even on return, just above!)
                # Falls thru on attempt and failure (Falsey)
            wrapper.ok		= False
            if default_cls is None and wrapper.rv:
                return wrapper.rv
            return default_cls() if default_cls else default_cls

        def reset():
            """Force an immediate attempt to run the underlying function, clearing any try count."""
            wrapper.ok		= True		# Start off assuming success (hence no initial delay)
            wrapper.cnt		= 0
            wrapper.lst		= None		# time of last function invocation

        wrapper.reset		= reset

        wrapper.reset()

        return wrapper

    return decorator


def ordinal( num ):
    ordinal_dict		= {1: "st", 2: "nd", 3: "rd"}
    q, mod			= divmod( num, 10 )
    suffix			= q % 10 != 1 and ordinal_dict.get(mod) or "th"
    return f"{num}{suffix}"


def commas( seq, final=None ):  # supply alternative final connector, eg. 'and', 'or'
    """Replace any numeric sequences eg. 1, 2, 3, 5, 7 w/ 1-3, 5 and 7.  Caller should
    usually sort numeric values before calling."""
    def int_seq( seq ):
        for i,iv in enumerate( seq[:-1] ):
            if type(iv) in (int,float):
                for j,jv in enumerate( seq[i:] ):
                    if type(jv) not in (int,float) or jv != iv + j:
                        j      -= 1
                        break
                if j > 1:
                    return (i,i+j)
        return None
    seq				= list( seq )
    while rng := int_seq( seq ):
        #print( f"int_seq: series {rng!r} ({seq[rng[0]]!r} - {seq[rng[1]]!r}) found in {seq!r}"  )
        beg			= seq[:rng[0]]
        nxt			= rng[1] + 1
        end			= seq[nxt:] if nxt < len( seq ) else []
        seq			= beg + [f"{seq[rng[0]]}-{seq[rng[1]]}"] + end
    if final and len(seq) > 1:
        seq			= seq[:-2] + [f"{seq[-2]} {final} {seq[-1]}"]
    return ', '.join( map( str, seq ))


def into_bytes( data: Union[bytes,str] ) -> bytes:
    """Convert hex data w/ optional '0x' prefix into bytes"""
    if isinstance( data, bytes ):
        return data
    if data[:2].lower() == '0x':
        data		= data[2:]
    return bytes.fromhex( data )


def round_onto( value, keys, keep_sign=True ):
    """Find the numeric key closest to value, maintaining sign if possible.  None and other
    non-numeric values are supported if they are present in keys.

    """
    if value in keys:
        return value
    keys			= sorted( k for k in keys if type(k) in (float,int) )
    near, near_i		= min(
        (abs( k - value), i)
        for i, k in enumerate( keys )
    )
    if keep_sign and (( value < 0 ) != ( keys[near_i] < 0 )):
        # The sign differs; if value -'ve but closest was +'ve, and there is a lower key available
        if value < 0:
            if near_i > 0:
                near_i	       -= 1
        else:
            if near_i + 1 < len( keys ):
                near_i	       += 1
    return keys[near_i]


entropy_signal_strengths	= {
    3: "very bad",
    2: "bad",
    1: "poor",
    0: "weak",
    -1: "ok",
    -2: "strong",
    None: "excellent",
}


def entropy_rating_dB( dB ):
    return entropy_signal_strengths[round_onto( dB, entropy_signal_strengths.keys(), keep_sign=True )]


def rate_dB( dB, what=None ):
    rating			= entropy_rating_dB( dB )
    result			= ''
    if what:
        result		       += f"{what}: "
    result		       += f"{dB:.1f}dB" if type( dB ) in (int,float) else f"{dB}"
    result		       += f" ({rating})"
    return result


def input_secure( prompt, secret=True, file=None ):
    """When getting secure (optionally secret) input from standard input, we don't want to use getpass, which
    attempts to read from /dev/tty.

    """
    if ( file or sys.stdin ).isatty():
        # From TTY; provide prompts, and do not echo secret input
        if secret:
            return getpass.getpass( prompt, stream=file )
        elif file:
            # Coming from some file; no prompt, read a line from the file source
            return file.readline()
        else:
            return input( prompt )
    else:
        # Not a TTY; don't litter pipeline output with prompts
        if file:
            return file.readline()
        return input()


def chunker( sequence, size ):
    while sequence:
        yield sequence[:size]
        sequence		= sequence[size:]


def hex_to_rgb( value, real=False, precision=4 ):
    """
    Convert hex color to ints, or reals rounded to a certain precision.

    >>> hex_to_rgb( "#ffffff" )
    (255, 255, 255)
    >>> hex_to_rgb( "#ffffffffffff" )
    (65535, 65535, 65535)
    >>> hex_to_rgb( "#ffffffffffff", real=True )
    (1.0, 1.0, 1.0)
    >>> hex_to_rgb( "#ffff80002000", real=True )
    (1.0, 0.5, 0.125)
    """
    nibble			= value.lstrip('#')
    nibs			= len( nibble )
    ints			= ( int( nibble[i:i + nibs // 3], 16 ) for i in range( 0, nibs, nibs // 3 ))
    return tuple( ( round( i / ( 16 ** ( nibs // 3 ) - 1 ), precision ) for i in ints ) if real else ints )


def rgb_to_hex( *rgb, bits=None, nibs=None ):
    """Convert int/real RGB values to hex.  For ints, assumes 8-bit values in range [0-255); for
    [0,65535), use bits=16.  For real-valued rgb in range (0,1), use bits=1.

    By default, will try to auto-select between bits=0/8/16, by examining rgb values; if all are <1,
    or <=1 and are float, then bits=0; otherwise if any are > 255, then bits=16.

    >>> rgb_to_hex( 255, 255, 255 )
    '#ffffff'
    >>> rgb_to_hex( 65535, 65535, 65535 )
    '#ffffffffffff'

    >>> rgb_to_hex( .5, .5, .5 )		# deduces (0,1)
    '#808080'

    >>> rgb_to_hex( .5, .75, .9, nibs=3  )	# same, but force 3 nibbles
    '#800bffe66'

    >>> rgb_to_hex( 11, 111, 1111 )		# deduces (0,65535]
    '#000b006f0457'

    >>> rgb_to_hex( 288, 255, 254, bits=8 )	# force 8-bit colors, w/ truncation
    '#fffffe'
    """
    assert len( rgb ) == 3, \
        "Requires RGB values"
    if bits is None:
        if all( c < 1 for c in rgb ) or all( c <= 1 and type( c ) is float for c in rgb ):
            bits		= 1
        elif any( c > 255 for c in rgb ):
            bits		= 16
        else:
            bits		= 8

    if nibs is None:
        nibs			= 2 if bits == 1 else bits // 4 + ( 1 if ( bits % 4 ) else 0 )

    end				= 2 ** bits - 1
    reals			= tuple( c / end for c in rgb )
    one				= 16 ** nibs - 1
    ints			= ( max( 0, min( one, int( round( r * one )))) for r in reals )

    return '#' + ''.join( "{i:0{nibs}x}".format( nibs=nibs, i=i ) for i in ints )


def hue_shift( color, shift=1/3 ):
    """Take a hex or (r,g,b) value, and shift its Hue "right" in the HSV spectrum by the given shift
    percentage (default: .1, or 10%).  Requires either '#aabbcc' hex colors, or real-values RGB in
    the range (0,1).

    >>> hue_shift( '#ff0000', shift=0 )
    '#ff0000'

    >>> hue_shift( '#ff0000', shift=1/3 )
    '#00ff00'

    >>> hue_shift( '#ff0000', shift=-1/3 )
    '#0000ff'

    >>> hue_shift( '#ff0000', shift=2/3 )
    '#0000ff'

    >>> hue_shift( '#ff0000', shift=1/2 )
    '#00ffff'

    >>> hue_shift( '#ff0000', shift=1/4 )
    '#80ff00'

    >>> hue_shift( '#ff0000', shift=1/10 )
    '#ff9900'

    >>> hue_shift( '#800000', shift=1/10 )
    '#804d00'

    >>> hue_shift( '#48c', shift=1/3 )
    '#cc4488'

    >>> hue_shift( (4/15, 8/15, 12/15), shift=-2/3 )  # Same RGB, but Hue shifted the other way 'round
    '#cc4488'
    """
    if isinstance( color, str ):
        rgb			= hex_to_rgb( color, real=1 )
    else:
        rgb			= tuple( color )
    assert len( rgb ) == 3 and all( 0.0 <= c <= 1.0 for c in rgb ), \
        f"Invalid RGB values for hue shift: {rgb!r}"
    h,s,v			= colorsys.rgb_to_hsv( *rgb )
    #log.warning( f"Color {color!r} == RGB {rgb!r} == HSV{(h,s,v)!r}" )
    h				= ( h + shift ) % 1.0  # if shift is -'ve, % (modulo) correctly shifts to +'ve
    rgb				= colorsys.hsv_to_rgb( h, s, v )
    return rgb_to_hex( *rgb )


def exponential_moving_average( current, sample, weight ):
    """exponential_moving_average -- rolling average without any data history

    Computes an exponential moving average:

        ( 1 - weight ) * current + weight * sample

    where the incoming sample has the given weight, and current samples have exponentially less
    influence on the current value.  Ignores a current value of None.

    """
    return sample if current is None else current + weight * ( sample - current )


def avg( seq ):
    vals			= list( seq )
    if vals:
        return sum( vals ) / len( vals )
    return math.nan


def rms( seq ):
    """Computes RMS for real/complex sequence of values"""
    if seq:
        return math.sqrt( avg( abs( s ) ** 2 for s in seq ))
    return 0


def is_power_of_2( n: int ) -> bool:
    return not ( n & ( n - 1 ))


class mixed_fraction( Fraction ):
    """A Fraction that represents whole multiples of itself as eg. 1+1/2 instead of 3/2"""
    def __str__( self ):
        whole, rest		= divmod( self.numerator, self.denominator )
        if whole and rest:
            return f"{whole}+{Fraction( rest, self.denominator)}"
        return super().__str__()


def remainder_after( proportions, scale=None, total=1 ):
    """Computes the sequence of what fraction must remain, after the preceding proportions have been
    removed.  Avoids scaling unless supplied and not falsey or 1.

    If the desired total to compare each fraction and the sum to isn't 1, supply it.  Also, an
    optional scaling factor for each fraction can be supplied (if the incoming stream of fractions
    don't sum to desired total).

    We support the proportions, scale and total being Fraction and integer values, and producing
    Fraction results.

    """
    f_total			= 0
    for f in proportions:
        if scale and scale != 1:
            f		       *= scale				# (0,total]
        f_starting		= total - f_total		# (0,total]
        f_removed		= f / f_starting		# (0,1]
        f_remaining		= total - total * f_removed     # (0,total]
        yield f_remaining
        f_total		       += f				# (0,total)


def fraction_allocated( reserves, scale=1 ):
    """From a sequence of remainder Fractions, compute the Fraction allocated at each point.

    If the reserves are known to be scaled by some factor, provide it (eg. if they are fixed-point
    fractions, provide the denominator as scale).

    Useful for computing the error between an original sequence of proportions, the resultant
    reserve Fractions, and the final proportion allocated to each party, when the reserve
    fraction is rounded (eg. represented as a fixed-point fraction).

        >>> proportions	= [ 57, 26, 103 ]
        >>> total = sum( proportions )
        >>> total
        186
        >>> reserves = list( remainder_after( proportions, scale=Fraction( 1, total )))
        >>> reserves
        [Fraction(43, 62), Fraction(103, 129), Fraction(0, 1)]
        >>> allocated = list( fraction_allocated( reserves ))
        >>> allocated
        [Fraction(19, 62), Fraction(13, 93), Fraction(103, 186)]

    """
    remaining			= Fraction( 1 )
    for reserve in reserves:
        remaining_after		= remaining * reserve / scale
        yield remaining - remaining_after
        remaining		= remaining_after


def uniq( seq, key=None ):
    """
    Removes duplicate elements from a sequence while preserving the order of the rest.

        >>> list(uniq([9,0,2,1,0]))
        [9, 0, 2, 1]

    The value of the optional `key` parameter should be a function that
    takes a single argument and returns a key to test the uniqueness.

        >>> list(uniq(["Foo", "foo", "bar"], key=lambda s: s.lower()))
        ['Foo', 'bar']
    """
    key				= key or (lambda x: x)
    seen			= set()
    for v in seq:
        k			= key( v )
        if k in seen:
            continue
        seen.add( k )
        yield v


def parse_scutil(input_data):
    """
    A simple parser for the custom 'scutil' data format where every object is a dictionary.
    """
    stack = []
    obj = None

    # Split lines and iterate through them
    for line in input_data.splitlines():
        log.debug( f"Parse: {line}" )
        line = line.strip()
        if not line:
            continue

        if line == '}':
            obj = stack.pop()
            continue

        *key,val = map(str.strip, line.split( ':' ))

        if val.startswith('<') and val.endswith('{'):
            # New dictionary/array detected
            val = {}
            if key:
                stack[-1][key[0]] = val
            stack.append(val)
            continue

        stack[-1][key[0]] = val

    assert not stack, \
        f"Invalid parse: {stack!r}"

    return obj



import requests
import re
import sys
import json
import logging
import traceback

from pathlib		import Path
from web3		import Web3

from .ethereum		import Chain, tokeninfos
from ..util		import commas

log				= logging.getLogger( "Tokens" )

top_100_url			= "https://etherscan.io/tokens?ps=100&p=1"


def scrapetokens( htmlfile ):
    """Scrape the top 100 ERC-20 tokens from Etherscan (or from such a file)."""

    if htmlfile:
        with open( htmlfile, 'r' ) as f:
            response_text	= f.read()
    else:
        # Probably won't work; Cloudflare scrape shielding?
        response		= requests.get(
            top_100_url,
            params	= dict(
                ps		= 100,
                p		= 1,
            )
        )
        assert response.status_code == 200, \
            f"Failed: {response.text}"
        response_text		= response.text
    # support anything with a link to [goerli.]etherscan.io/{token,address}/0x...
    for token_m in re.finditer( r'(?:img src="([^"]+)".{0,100})?https://(?:[a-z]+.)?etherscan.io/token/(0x[0-9a-fA-F]+)', response_text ):
        yield token_m.groups()


if __name__ == "__main__":
    # Look at any .json files specified, for a list of ERC-20 token details, or __file__.json
    # Save any token images found in Tokens/{symbol}_<size>.wepb
    erc20s			= {}
    chain, argv			= sys.argv[1],sys.argv[2:]
    chain,			= ( c for c in Chain if c.name.lower() == chain.lower() )

    # See if any .json files are specified to use; otherwise, default to eg. ./Tokens-Ethereum.json
    argv_json			= list( filter( lambda n: n.endswith( '.json' ), argv ))
    if not argv_json:
        here_json		= Path( str( Path( __file__ ).with_suffix( '' )) + f"-{chain.name}.json" )
        if here_json.exists():
            argv_json		= [ here_json ]
    here_icon			= Path( __file__ ).with_suffix( '' )
    log.warning( f"Icon path {here_icon} exists: {here_icon.exists()}" )

    for erc20sfile in argv_json:
        with open( erc20sfile, 'r' ) as f:
            for t_i in json.loads( f.read() ):
                erc20s[t_i['address']] = t_i
    log.warning( f"Loaded {len(erc20s)} known ERC-20 tokens from {len(argv_json)} JSON files" )

    # Look at any remaining files specified, as HTML containing etherescan.io/token/... addresses
    argv_html			= list( filter( lambda n: n not in argv_json, argv ))
    erc20s_n			= {}
    for htmlfile in argv_html:
        htmlpath		= Path( htmlfile )
        for image,token in scrapetokens( htmlpath ):
            # See if we've already seen the token; otherwise, get its info
            try:
                t		= Web3.to_checksum_address( token )
                if t not in erc20s:
                    erc20s_n[t], = tokeninfos( t, chain=chain )
                    log.warning( f"Found {erc20s_n[t]['symbol']:8}: {t}" )
            except Exception as exc:
                log.warning( f"Failed to query {token}: {exc}; {traceback.format_exc()}" )
                continue

            # See if we've already got its image; otherwise, collect it.
            info		= erc20s.get( t, erc20s_n.get( t ))
            symbol		= info['symbol']
            if here_icon.exists() and image:
                log.warning( f"Found image: {image}" )
                found_icon	= htmlpath.resolve().parent / Path( image )
                found_suff	= re.match( r".*((?:\d+)?[.].*)$", image )
                if found_suff:
                    found_suff	= found_suff.group( 1 )
                    if not found_suff.startswith( '.' ):
                        found_suff = '_'+found_suff  # w/ a size eg. 28.webp, use _28.webp
                log.warning( f"Found suffix: {found_suff}" )
                if found_icon.exists() and found_suff:
                    saved_icon	= here_icon / f"{symbol}{found_suff}"
                    log.warning( f"Found {symbol:8} token image file: {found_icon} w/ suffix: {found_suff}; moving to {saved_icon}" )
                    found_icon.rename( saved_icon )
                    info['icon'] = str( Path( *saved_icon.parts[-2:] ))

    log.warning( f"Loaded {len(erc20s_n)} new ERC-20 tokens from {len(argv_html)} HTML files" )

    erc20s_t                    = erc20s | erc20s_n
    log.warning( f"Summed {len(erc20s_t)} ERC-20 tokens" )

    # Kick out any without Ascii symbol names
    erc20s_eject		= []
    for a,i in erc20s_t.items():
        if not all( 32 < ord( c ) < 128 for c in i['symbol'] ):
            erc20s_eject.append( a )
    if erc20s_eject:
        log.warning( f"Ejecting {len(erc20s_eject)} tokens w/ bad symbols: {commas( erc20s_t[a]['name'] for a in erc20s_eject )}" )
        for a in erc20s_eject:
            del erc20s_t[a]
        log.warning( f"Left w/{len(erc20s_t)} ERC-20 tokens" )

    try:
        erc20s_list             = sorted( list( erc20s_t.values() ), key=lambda i: i['name'] )
    except Exception as exc:
        print( json.dumps( list( erc20s_t.values()), indent=4, sort_keys=True ))
        log.error( f"Failed to sort: ERC-20s list may be incorrect: {exc}" )
    else:
        print( json.dumps( erc20s_list, indent=4, sort_keys=True ))




#
# Python-slip39 -- Ethereum SLIP-39 Account Generation and Recovery
#
# Copyright (c) 2022, Dominion Research & Development Corp.
#
# Python-slip39 is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.  It is also available under alternative (eg. Commercial) licenses, at
# your option.  See the LICENSE file at the top of the source tree.
#
# Python-slip39 is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#

from __future__		import annotations

import argparse
import codecs
import logging

from .			import Account
from .api		import random_secret
from .util		import log_cfg, log_level, input_secure
from .layout		import write_pdfs
from .defaults		import (   # noqa: F401
    CARD, CARD_SIZES, PAPER, WALLET, WALLET_SIZES,
    BITS, BITS_DEFAULT,
    FILENAME_FORMAT,
    FILENAME_KEYWORDS,
    CRYPTO_PATHS,
)

__author__                      = "Perry Kundert"
__email__                       = "perry@dominionrnd.com"
__copyright__                   = "Copyright (c) 2022 Dominion Research & Development Corp."
__license__                     = "Dual License: GPLv3 (or later) and Commercial (see LICENSE)"

log				= logging.getLogger( __package__ )


def main( argv=None ):
    ap				= argparse.ArgumentParser(
        description = "Create and output SLIP-39 encoded Seeds and Paper Wallets to a PDF file.",
        epilog = "" )
    ap.add_argument( '-v', '--verbose', action="count",
                     default=0,
                     help="Display logging information." )
    ap.add_argument( '-q', '--quiet', action="count",
                     default=0,
                     help="Reduce logging output." )
    ap.add_argument( '-o', '--output',
                     default=FILENAME_FORMAT,
                     help=f"Output PDF to file or '-' (stdout); formatting w/ {', '.join( FILENAME_KEYWORDS )} allowed" )
    ap.add_argument( '-t', '--threshold',
                     default=None,
                     help="Number of groups required for recovery (default: half of groups, rounded up)" )
    ap.add_argument( '-g', '--group', action='append',
                     help="A group name[[<require>/]<size>] (default: <size> = 1, <require> = half of <size>, rounded up, eg. 'Frens(3/5)' )." )
    ap.add_argument( '-f', '--format', action='append',
                     default=[],
                     help=f"Specify crypto address formats: {', '.join( Account.FORMATS )}; default: " + ', '.join(
                         f'{c}:{Account.address_format(c)}' for c in Account.CRYPTO_NAMES.values()
                     ))
    ap.add_argument( '-c', '--cryptocurrency', action='append',
                     default=[],
                     help="A crypto name and optional derivation path (eg. '../<range>/<range>'); defaults: " + ', '.join(
                         f"{c}:{Account.path_default(c)}" for c in Account.CRYPTO_NAMES.values()
                     ))
    ap.add_argument( '-p', '--path',
                     default=None,
                     help="Modify all derivation paths by replacing the final segment(s) w/ the supplied range(s), eg. '.../1/-' means .../1/[0,...)")
    ap.add_argument( '-j', '--json',
                     default=None,
                     help="Save an encrypted JSON wallet for each Ethereum address w/ this password, '-' reads it from stdin (default: None)" )
    ap.add_argument( '-w', '--wallet',
                     default=None,
                     help="Produce paper wallets in output PDF; each wallet private key is encrypted this password (use --wallet=\"\" for empty password)" )
    ap.add_argument( '--wallet-hint',
                     default=None,
                     help="Paper wallets password hint" )
    ap.add_argument( '--wallet-format',
                     default=None,
                     help=f"Paper wallet size; {', '.join(WALLET_SIZES.keys())} or '(<h>,<w>),<margin>' (default: {WALLET})" )
    ap.add_argument( '-s', '--secret',
                     default=None,
                     help="Use the supplied BIP-39 Mnemonic or 128-, 256- or 512-bit hex value as the secret seed; '-' reads it from stdin (eg. output from slip39.recover)" )
    ap.add_argument( '--bits',
                     default=None,  # Do not enforce default of 128-bit seeds
                     help=f"Ensure that the seed is of the specified bit length; {', '.join( map( str, BITS ))} supported." )
    ap.add_argument( '--using-bip39', action='store_true',
                     default=None,
                     help="Generate Seed from secret Entropy using BIP-39 generation algorithm (encode as BIP-39 Mnemonics, encrypted using --passphrase)" )
    ap.add_argument( '--passphrase',
                     default=None,
                     help="Encrypt the master secret w/ this passphrase, '-' reads it from stdin (default: None/'')" )
    ap.add_argument( '-C', '--card',
                     default=None,
                     help=f"Card size; {', '.join(CARD_SIZES.keys())} or '(<h>,<w>),<margin>' (default: {CARD})" )
    ap.add_argument( '--no-card', dest="card", action='store_false',
                     help="Disable PDF SLIP-39 mnemonic card output" )
    ap.add_argument( '--paper',
                     default=None,
                     help=f"Paper size (default: {PAPER})" )
    ap.add_argument( '--cover', dest="cover_page", action='store_true',
                     default=True,
                     help="Produce PDF SLIP-39 cover page" )
    ap.add_argument( '--no-cover', dest="cover_page", action='store_false',
                     help="Disable PDF SLIP-39 cover page" )
    ap.add_argument( '--text', action='store_true',
                     default=None,
                     help="Enable textual SLIP-39 mnemonic output to stdout" )
    ap.add_argument( '--watermark',
                     default=None,
                     help="Include a watermark on the output SLIP-39 mnemonic cards" )
    ap.add_argument( '--double-sided', action='store_true',
                     default=None,
                     help="Enable double-sided PDF (default)" )
    ap.add_argument( '--no-double-sided', dest="double_sided", action='store_false',
                     help="Disable double-sided PDF" )
    ap.add_argument( '--single-sided', dest="double_sided", action='store_false',
                     help="Enable single-sided PDF" )
    ap.add_argument( 'names', nargs="*",
                     help="Account names to produce; if --secret Entropy is supplied, only one is allowed.")
    args			= ap.parse_args( argv )

    # Set up logging; also, handle the degenerate case where logging has *already* been set up (and
    # basicConfig is a NO-OP), by (also) setting the logging level
    log_cfg['level']		= log_level( args.verbose - args.quiet )
    logging.basicConfig( **log_cfg )
    if args.verbose:
        logging.getLogger().setLevel( log_cfg['level'] )

    # Confirm sanity of args
    log.debug( f"args: {args!r}" )
    if args.path:
        assert args.path.startswith( 'm/' ) or ( args.path.startswith( '..' ) and args.path.lstrip( '.' ).startswith( '/' )), \
            "A --path must start with 'm/', or '../', indicating intent to replace 1 or more trailing components of each cryptocurrency's derivation path"

    # If any --format <crypto>:<format> address formats provided
    for cf in args.format:
        try:
            Account.address_format( *cf.split( ':' ) )
        except Exception as exc:
            log.error( f"Invalid address format: {cf}: {exc}" )
            raise

    bits_desired		= int( args.bits ) if args.bits else BITS_DEFAULT

    master_secret		= args.secret
    if master_secret:
        # Master secret seed may be supplied as hex or a BIP-39 Mnemonic (leave as str)
        if master_secret == '-':
            master_secret	= input_secure( 'Master secret (hex or BIP-39): ', secret=True )
        else:
            log.warning( "It is recommended to not use '-s|--secret <hex>'; specify '-' to read from input" )
        if master_secret.lower().startswith('0x'):
            master_secret	= master_secret[2:]
        if all( c in "0123456789abcdef" for c in master_secret.lower() ):
            master_secret	= codecs.decode( master_secret, 'hex_codec' )
    else:
        # Generate a random secret seed, as bytes
        master_secret		= random_secret( bits_desired // 8 )
    if isinstance( master_secret, bytes ):
        master_secret_bits		= len( master_secret ) * 8
        if master_secret_bits not in BITS:
            raise ValueError( f"A {master_secret_bits}-bit master secret was supplied; One of {BITS!r} expected" )
        if args.bits and master_secret_bits != bits_desired:  # If a certain seed size specified, enforce
            raise ValueError( f"A {master_secret_bits}-bit master secret was supplied, but {bits_desired} bits was specified" )

    # SLIP-39 Passphrase.  This is not recommended, as A) it is another thing that must be saved in
    # addition to the SLIP-39 Mnemonics in order to recover the Seed, and B) it is not implemented
    # by the Trezor hardware wallet.  Also used for --bip39, where the wallets are produced from the
    # master_secret Entropy using BIP-39 standard Seed generation.
    passphrase			= args.passphrase or ""
    if passphrase == '-':
        passphrase		= input_secure( 'Master seed passphrase: ', secret=True )
    elif passphrase:
        log.warning( "It is recommended to not use '-p|--passphrase <password>'; specify '-' to read from input" )
    if passphrase:
        if args.using_bip39 or isinstance( master_secret, str ):
            log.warning( "The BIP-39 Passphrase will be used only for deducing Crypto addresses; you must remember it, for BIP-39 wallet recovery" )
        else:
            log.warning( "The SLIP-39 Standard Passphrase is not compatible w/ the Trezor hardware wallet; use its 'Hidden wallet' feature instead" )

    # Optional Paper Wallet and/or JSON Wallet file passwords
    wallet_pwd			= args.wallet
    if wallet_pwd:
        if wallet_pwd == '-':
            wallet_pwd		= input_secure( 'Paper Wallet password: ', secret=True )
        else:
            log.warning( "It is recommended to not use '-w|--wallet <password>'; specify '-' to read from input" )
    wallet_pwd_hint		= args.wallet_hint
    wallet_format		= args.wallet_format

    json_pwd			= args.json
    if json_pwd is not None:
        if json_pwd == '-':
            json_pwd		= input_secure( 'Ethereum JSON wallet file password: ', secret=True )
        else:
            log.warning( "It is recommended to not use '-j|--json <password>'; specify '-' to read from input" )

    try:
        # Output the filenames of the emitted PDFs, one per line.
        print( "\n".join( write_pdfs(
            names		= args.names,
            master_secret	= master_secret,
            passphrase		= passphrase,
            using_bip39		= args.using_bip39,
            group		= args.group,
            group_threshold	= args.threshold,
            cryptocurrency	= args.cryptocurrency,
            edit		= args.path,
            card_format		= args.card,    # False inhibits SLIP-39 Card output
            paper_format	= args.paper,
            filename		= args.output,  # outputs to the current working dir, by default
            json_pwd		= json_pwd,
            text		= args.text,
            wallet_pwd		= wallet_pwd,
            wallet_pwd_hint	= wallet_pwd_hint,
            wallet_format	= wallet_format,
            cover_page		= args.cover_page,
            watermark		= args.watermark,
            double_sided	= args.double_sided,
        )))
    except Exception as exc:
        log.exception( f"Failed to write PDFs: {exc}" )
        return 1
    return 0




#
# Python-slip39 -- Ethereum SLIP-39 Account Generation and Recovery
#
# Copyright (c) 2022, Dominion Research & Development Corp.
#
# Python-slip39 is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.  It is also available under alternative (eg. Commercial) licenses, at
# your option.  See the LICENSE file at the top of the source tree.
#
# Python-slip39 is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#

from __future__          import annotations

from .			import cli

__author__                      = "Perry Kundert"
__email__                       = "perry@dominionrnd.com"
__copyright__                   = "Copyright (c) 2022 Dominion Research & Development Corp."
__license__                     = "Dual License: GPLv3 (or later) and Commercial (see LICENSE)"

cli()




#
# Python-slip39 -- Ethereum SLIP-39 Account Generation and Recovery
#
# Copyright (c) 2022, Dominion Research & Development Corp.
#
# Python-slip39 is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.  It is also available under alternative (eg. Commercial) licenses, at
# your option.  See the LICENSE file at the top of the source tree.
#
# Python-slip39 is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
from __future__          import annotations

import ast
import io
import json
import logging
import math
import os
import warnings

from datetime		import datetime
from collections	import namedtuple, defaultdict
from collections.abc	import Callable
from pathlib		import Path
from typing		import Dict, List, Tuple, Optional, Sequence, Any

import qrcode
import qrcode.image.svg
import fpdf		# FPDF, FlexTemplate, FPDF_FONT_DIR
import fpdf.svg

from ..api		import Account, cryptopaths_parser, create, enumerate_mnemonic, group_parser
from ..util		import chunker
from ..recovery		import recover, produce_bip39
from ..defaults		import (
    FONTS, CARD, CARD_SIZES, PAPER, ORIENTATION, PAGE_MARGIN, MM_IN, PT_IN,
    WALLET, WALLET_SIZES,
    GROUPS, GROUP_THRESHOLD_RATIO,
    FILENAME_FORMAT,
)
from .components	import (
    Coordinate, Region, Box, Image, Text,
    page_dimensions, layout_components, layout_card, layout_wallet,
)
from .printer		import (
    printer_output,
)

__author__                      = "Perry Kundert"
__email__                       = "perry@dominionrnd.com"
__copyright__                   = "Copyright (c) 2022 Dominion Research & Development Corp."
__license__                     = "Dual License: GPLv3 (or later) and Commercial (see LICENSE)"

# Optionally support output of encrypted JSON files
eth_account			= None
try:
    import eth_account
except ImportError:
    pass

log				= logging.getLogger( __package__ )


class FPDF_Autoload_Fonts( fpdf.FPDF ):
    """If attempting to set an unknown font, tries to load it. """
    def __init__( self, *args, font_dir=None, **kwds ):
        self.font_dir		= font_dir
        if self.font_dir is None:
            self.font_dir	= font_dir if font_dir is not None else Path( __file__ ).resolve().parent / "font"
        return super().__init__( *args, **kwds )

    def set_font( self, family=None, style="", size=0 ):
        # Deduce a set of lists of features synonyms, and expected fontkey
        feature_suffixes	= dict(
            B	= ( 'bold', ),
            I	= ( 'italic', 'oblique' ),      # Typical aliases seen
            U	= (),				# Ignored for font loading purposes
        )

        assert all( c in feature_suffixes for c in style ), \
            f"Unrecognized font style in {style!r}"
        features		= set( feature_suffixes[c] for c in style )
        if not features:
            features.add( ( 'regular', ) )
        family			= family.lower()
        fontkey			= family + style

        try:
            return super().set_font( family=family, style=style, size=size )
        except fpdf.FPDFException as exc:
            if "Undefined font" not in str( exc ):
                raise
            log.info( f"Loading custom font {fontkey}: {exc}" )

        # Search for the *shortest* fname that matches the requested family/features.  A font with
        # no features suffix will be assumed to have the -Regular suffix.
        fname_best		= None
        for fname in sorted( self.font_dir.glob( '*.ttf' ) ):
            log.debug( f"Evaluating {family=} + {features=} against:  {fname}" )
            # Eg "DejaVuSansMono-BoldOblique.ttf" --> "dejavusansmono", "boldoblique"
            fname_segments	= fname.stem.split( '-', 2 )		# the file name w/o path or suffix
            fname_family	= fname_segments[0].lower()
            fname_features 	= ''.join( fname_segments[1:] ).lower() or 'regular'
            if family not in fname_family:
                continue
            # Family matches; now, verify that the features eg. Italic/Oblique, Bold, Regular are in
            # the fname_features.
            for feature in features:
                if not any( synomym in fname_features for synomym in feature ):
                    break
            else:
                # Every one of features specified (or one of the feature's synonyms) appeared in
                # this fname's feature segment.  It is a candidate; is it better (shorter) than the
                # one currently found?  Eg. Font-Bold.ttf better than Font-Oblique.ttf, if no
                # Font.ttf or Font-Regular.ttf
                if fname_best is None or len( fname.stem ) < len( fname_best.stem ):
                    log.debug( f"Font named {family=} + {features=} matching: {fname} and better than {fname_best}" )
                    fname_best	= fname
                else:
                    log.debug( f"Font named {family=} + {features=} matching: {fname}, but worse than  {fname_best}" )
                continue
            log.debug( f"Font named {family=} + {features=} no match: {fname}" )
        if fname_best:
            log.info( f"Font named {family=} + {features=} loading:  {fname_best} (as '{family}{style}')" )
            self.add_font( family=family, style=style, fname=str( fname_best ))
        else:
            log.warning( f"Font name {family=} and {features=} not found for: {fontkey}" )
            raise fpdf.FPDFException( f"TTF Font file not found for: {fontkey}" )

        return super().set_font( family=family, style=style, size=size )


LayoutPDF			= namedtuple( 'LayoutPDF', [
    'comps_pp', 'orientation', 'page_xy', 'pdf', 'comp_dim',
] )


def layout_pdf(
    comp_dim: Optional[Coordinate] = None,      # In mm.; Default: full working page size
    page_margin_mm: Optional[float] = None,     # Default: 1/4" in mm.
    orientation: Optional[str]	= None,		# Default: portrait
    paper_format: Optional[Any]	= None,         # Can be a paper name (Letter) or (x, y) dimensions in mm.
    font_dir: Optional[Path]	= None,
) -> Tuple[int, str, Callable[[int],[int, Coordinate]], fpdf.FPDF, Coordinate]:
    """Find the ideal orientation for the most components of dimensions 'comp_dim'.  Returns the
    number of components per page, the FPDF, and a function useful for laying out templates on the
    pages of the PDF.

    Returns the details required to start rendering components to the FPDF, including the
    component dimensions, 'comp_dim' in mm.

    """
    pdf				= FPDF_Autoload_Fonts(
        orientation	= orientation or ORIENTATION,
        format		= paper_format or PAPER,
    )
    pdf.set_margin( 0 )

    # Establish page dimensions, net margins; if no component dimension, default to page dimensions
    if page_margin_mm is None:
        page_margin_mm	= PAGE_MARGIN * MM_IN
    page_dim		= page_dimensions( pdf, page_margin_mm=page_margin_mm )
    if comp_dim is None:
        # Default to 1 component per page (the user must limit themselves to the page size)
        comp_dim	= page_dim

    # Lay out components on page, computing the number of components that will fit
    comps_pp,page_xy	= layout_components(
        pdf,
        comp_dim	= comp_dim,
        page_margin_mm	= page_margin_mm
    )
    log.info( f'PDF Layout: {comps_pp} components of {comp_dim.x / MM_IN:7.3f}" x {comp_dim.y / MM_IN:7.3f}"'
              f' on an {page_dimensions( pdf ).x / MM_IN:7.3f}" x {page_dimensions( pdf ).y / MM_IN:7.3f}" {orientation or ORIENTATION} page' )
    return LayoutPDF( comps_pp, orientation, page_xy, pdf, comp_dim )


def output_pdf( *args, **kwds ):
    warnings.warn(
        "output_pdf() is deprecated; use produce_pdf instead.",
        PendingDeprecationWarning,
    )
    _,pdf,accounts		= produce_pdf( *args, **kwds )
    return pdf,accounts


def produce_pdf(
    name: str,
    group_threshold: int,			# SLIP-39 Group Threshold required
    groups: Dict[str,Tuple[int,List[str]]],     # SLIP-39 Groups {<name>: (<need>,[<mnemonic>,...])
    accounts: Sequence[Sequence[Account]],      # The crypto account(s); at least 1 of each required
    using_bip39: bool,				# Using BIP-39 wallets Seed generation
    card_format: str		= CARD,		# 'index' or '(<h>,<w>),<margin>'
    paper_format: Any		= None,		# 'Letter', (x,y) dimensions in mm.
    orientations: Sequence[str]	= None,		# available orientations; default portrait, landscape
    cover_text: Optional[str]	= None,		# Any Cover Page text; we'll append BIP-39 if 'using_bip39'
    watermark: Optional[str]	= None,
    double_sided: Optional[bool]= None,
) -> Tuple[Tuple[str,str], fpdf.FPDF, Sequence[Sequence[Account]]]:
    """Produces an FPDF containing the specified SLIP-39 Mnemonics group recovery cards.

    Returns the required Paper description [<format>,<orientation>], the FPDF containing the
    produced cards, and the cryptocurrency accounts from the supplied slip39.Details.

    Makes available several fonts.

    Produces double-sided PDF by default, containing QR codes for SLIP-39 Mnemonics on the back.
    """
    if paper_format is None:
        paper_format		= PAPER
    if orientations is None:
        orientations		= ('portrait', 'landscape')

    double_sided		= True if double_sided is None else bool( double_sided )

    # Deduce the card size.  All papers sizes are specified in inches.
    try:
        (card_h,card_w),card_margin = CARD_SIZES[card_format.lower()]
    except KeyError:
        (card_h,card_w),card_margin = ast.literal_eval( card_format )
    card_size			= Coordinate( y=card_h, x=card_w )

    # Compute how many cards per page.  Flip page portrait/landscape to match the cards'.  Use the length
    # of the first Group's first Mnemonic to determine the number of mnemonics on each card.
    num_mnemonics		= len( groups[next(iter(groups.keys()))][1][0].split() )
    card			= layout_card(  # layouts are computed in inches.
        card_size, card_margin, num_mnemonics=num_mnemonics )
    card_dim			= card.mm()     # converts to mm for PDF

    # Find the best PDF and orientation, by max of returned cards_pp (cards per page).  Assumes
    # layout_pdf returns a tuple that can be compared; cards_pp,orientation,... will always sort.
    page_margin_mm		= PAGE_MARGIN * MM_IN
    cards_pp,orientation,page_xy,pdf,_ = max(
        layout_pdf( card_dim, page_margin_mm, orientation=orientation, paper_format=paper_format )
        for orientation in orientations
    )
    log.debug( f"Page: {paper_format} {orientation} {pdf.epw:.8}mm w x {pdf.eph:.8}mm h w/ {page_margin_mm}mm margins,"
               f" Card: {card_format} {card_dim.x:.8}mm w x {card_dim.y:.8}mm h == {cards_pp} cards/page" )

    card_elements		= list( card.elements() )
    if log.isEnabledFor( logging.DEBUG ):
        log.debug( f"Card elements: {json.dumps( card_elements, indent=4)}" )
    tpl				= fpdf.FlexTemplate( pdf, card_elements )

    group_reqs			= list(
        f"{g_nam}({g_of}/{len(g_mns)})" if g_of != len(g_mns) else f"{g_nam}({g_of})"
        for g_nam,(g_of,g_mns) in groups.items() )
    requires			= f"Recover w/ {group_threshold} of {len(group_reqs)} groups {', '.join(group_reqs[:4])}{'...' if len(group_reqs)>4 else ''}"

    # Convert all of the first group's account(s) to an address QR code
    assert accounts and accounts[0], \
        "At least one cryptocurrency account must be supplied"
    qr_acct			= {}
    for i,acct in enumerate( accounts[0] ):
        qrc			= qrcode.QRCode(
            version		= None,
            error_correction	= qrcode.constants.ERROR_CORRECT_M,
            box_size		= 10,
            border		= 1,
            image_factory	= qrcode.image.svg.SvgPathImage,
        )
        qrc.add_data( acct.address )
        qrc.make( fit=True )

        qr_acct[i]		= qrc.make_image()
        if log.isEnabledFor( logging.INFO ):
            f			= io.StringIO()
            qrc.print_ascii( out=f )
            f.seek( 0 )
            for line in f:
                log.info( line.strip() )

    if cover_text:
        # Cover page is always laid out in landscape, and rotated if necessary
        cvw			= max( pdf.epw, pdf.eph )
        cvh			= min( pdf.epw, pdf.eph )
        cover			= Region(
            'cover', 0, 0, cvw / MM_IN, cvh / MM_IN
        ).add_region_relative(
            Box( 'cover-interior', x1=+1/2, y1=+1/2, x2=-1/2, y2=-1/2 )
        )
        cover.add_region_proportional(
            Image( 'cover-image', x1=1/4, x2=3/4, y1=1/4, y2=3/4, priority=-3 )
        ).square().add_region(
            Image( 'cover-fade', priority=-2 )
        )
        cover.add_region_proportional(
            Text( 'cover-text', y2=1/45, font='mono', multiline=True )  # 1st line
        )
        cover.add_region_proportional(
            Region( 'cover-rhs', x1=3/5, y1=1/5 )  # On right, below full-width header
        ).add_region_proportional(
            Text( 'cover-sent', y2=1/25, font='mono', multiline=True )  # 1st line
        )

        cover_elements		= list( cover.elements() )
        if log.isEnabledFor( logging.DEBUG ):
            log.debug( f"Cover elements: {json.dumps( cover_elements, indent=4)}" )
        tpl_cover		= fpdf.FlexTemplate( pdf, cover_elements )
        images			= os.path.dirname( __file__ )
        tpl_cover['cover-image'] = os.path.join( images, 'SLIP-39.png' )
        tpl_cover['cover-fade'] = os.path.join( images, '1x1-ffffffbf.png' )

        slip39_mnems		= []
        slip39_group		= []
        g_nam_max		= max( map( len, groups.keys() ))
        for g_nam,(g_of,g_mns) in groups.items():
            slip39_mnems.extend( g_mns )
            slip39_group.append( f"{g_nam:{g_nam_max}}: {g_of} of {len(g_mns)} to recover" )
            slip39_group.extend( f"  {i+1:2}: ____________________" for i in range( len( g_mns )))
        if using_bip39:
            # Add the BIP-39 Mnemonics to the cover_text, by recovering the master_secret from the
            # SLIP-39 Mnemonics.
            master_secret	= recover( slip39_mnems )
            bip39_enum		= enumerate_mnemonic( produce_bip39( entropy=master_secret ) )
            rows		= ( len( bip39_enum ) + 4 ) // 5
            cols		= ( len( bip39_enum ) + rows - 1 ) // rows
            cover_text	       += "\n--------------8<-------------[ cut here ]------------------"
            cover_text	       += f"\n\n            Your {len(bip39_enum)}-word BIP-39 Seed Phrase is:\n"
            for r in range( rows ):
                cover_text     += "\n    "
                for c in range( cols ):
                    word	= bip39_enum.get( c * rows + r )
                    cover_text += f"{word or '':12}"
            cover_text	       += "\n"

        tpl_cover['cover-text']	= cover_text
        cover_sent		= "SLIP-39 Mnemonic Card Recipients:\n\n"
        cover_sent	       += "\n".join( slip39_group )
        tpl_cover['cover-sent']	= cover_sent

        pdf.add_page()
        if orientation.lower().startswith( 'p' ):
            tpl_cover.render( offsetx=pdf.epw, rotate=-90.0 )
        else:
            tpl_cover.render()

        if double_sided:
            pdf.add_page()


    # Compute the contents of the cards; the keys and their values are the attributes of
    # each template.  Creates pages of cards (<pos>,<front>,<back>).
    page			= []  # A sequence of pages [[<card>,..],..]
    card_n			= 0
    for g_n,(g_name,(g_of,g_mnems)) in enumerate( groups.items() ):
        for mn_n,mnem in enumerate( g_mnems ):
            p_n,(p_x,p_y)	= page_xy( card_n )
            card_n	       += 1
            if p_n == len(page):
                page.append([])
            page[-1].append( ((p_n,(p_x,p_y)), dict(), dict()) )

            _,f,b		= page[-1][-1]

            f['card-title']	= \
            b['card-title']	= f"SLIP39 {g_name}({mn_n+1}/{len(g_mnems)}) for: {name}"
            f['card-requires']	= requires
            f['card-crypto1']	= f"{accounts[0][0].crypto} {accounts[0][0].path}: {accounts[0][0].address}"
            f['card-qr1']	= io.BytesIO( qr_acct[0].to_string(encoding='unicode').encode('UTF-8')) # get_image()
            if len( accounts[0] ) > 1:
                f['card-crypto2'] = f"{accounts[0][1].crypto} {accounts[0][1].path}: {accounts[0][1].address}"
                f['card-qr2']	= io.BytesIO( qr_acct[1].to_string(encoding='unicode').encode('UTF-8'))  # get_image()
            f[f'card-g{g_n}']	= \
            b[f'card-g{g_n}']	= f"{g_name:7.7}..{mn_n+1}" if len(g_name) > 8 else f"{g_name} {mn_n+1}"
            if watermark:
                f['card-watermark'] = watermark
            for n,m in enumerate_mnemonic( mnem ).items():
                f[f"mnem-{n}"]	= m

            # Back contains QR code of the card's SLIP-39 Mnemonic
            qrc			= qrcode.QRCode(
                version		= None,
                error_correction = qrcode.constants.ERROR_CORRECT_M,
                box_size	= 10,
                border		= 1,
                image_factory	= qrcode.image.svg.SvgPathImage
            )
            qrc.add_data( mnem )
            qrc.make( fit=True )

            b['card-qr-mnem']	= io.BytesIO( qrc.make_image().to_string(encoding='unicode').encode('UTF-8'))  # get_image()

    # Now render the front and (optionally) back cards for each page
    for p_n,p in enumerate( page ):
        pdf.add_page()
        for c_n,((_,(p_x,p_y)),f,b) in enumerate( p ):
            for key,txt in f.items():
                tpl[key] 	= txt
            tpl.render( offsetx=p_x, offsety=p_y )
        if not double_sided:
            continue
        pdf.add_page()
        for c_n,((_,(p_x,p_y)),f,b) in enumerate( p ):
            for key,txt in b.items():
                tpl[key] 	= txt
            tpl.render( offsetx=pdf.epw - p_x - card_dim.x, offsety=p_y )

    return (paper_format,orientation),pdf,accounts


def write_pdfs(
    names		= None,		# sequence of [ <name>, ... ], or { "<name>": <details> }
    master_secret	= None,		# Derive SLIP-39 details from this seed (if not supplied in names)
    passphrase		= None,		# UTF-8 encoded passphrase; default is '' (empty)
    using_bip39		= False,        # Generate Seed from Entropy via BIP-39 generation algorithm
    group		= None,		# Group specifications, [ "Frens(3/6)", ... ]
    group_threshold	= None,		# int, or 1/2 of groups by default
    cryptocurrency	= None,		# sequence of [ 'ETH:<path>', ... ] to produce accounts for
    edit		= None,		# Adjust crypto paths according to the provided path edit
    card_format		= None,		# Eg. "credit"; False outputs no SLIP-39 Mnemonic cards to PDF
    paper_format	= None,		# Eg. "Letter", "Legal", (x,y)
    filename		= None,		# A file name format w/ {path}, etc. formatters
    filepath		= True,		# A file path, if PDF output to file is desired; ''/True implies current dir.
    printer		= None,		# A printer name (or True for default), if output to printer is desired
    json_pwd		= None,		# If JSON wallet output desired, supply password
    text		= None,		# Truthy outputs SLIP-39 recover phrases to stdout
    wallet_pwd		= None,		# If paper wallet images desired, supply password
    wallet_pwd_hint	= None,		# an optional password hint
    wallet_format	= None,		# and a paper wallet format (eg. 'quarter')
    wallet_paper	= None,		# default Wallets to output on Letter format paper,
    cover_page		= True,         # Produce a cover page (including BIP-39 Mnemonic, if using_bip39)
    watermark		= None,		# Any watermark desired on each output
    double_sided	= None,
):
    """Writes a PDF containing a unique SLIP-39 encoded Seed Entropy for each of the names specified.

    If a sequence of multiple 'names' are supplied, then no master_secret is allowed (we must
    generate for each name), or a dict of { 'name': <details>, ... } for each name must be provided
    (neither master_secret nor using_bip39 allowed).

    If we are generating master_secrets, and accounts, a number of Cryptocurrency wallet public
    addresses and QR codes are generated; optionally, 'using_bip39' to force BIP-39 standard Seed
    generation (instead of SLIP-39 standard, which uses the Seed Entropy directly).

    If 'using_bip39' and 'cover_page', the BIP-39 Mnemonic phrase card is also produced.  It is
    recommended to destroy this BIP-39 Mnemonic (ideally), or store it very, very securely (not
    recommended).  Use the SLIP-39 "Recover" Controls, instead, to recover the BIP-39 Mnemonic
    phrase when needed to restore a hardware wallet.

    Returns a { "<filename>": <details>, ... } dictionary of all PDF files written, and each of
    their account details.

    """
    # Convert sequence of group specifications into standard { "<group>": (<needs>, <size>) ... }
    if isinstance( group, dict ):
        groups			= group
    else:
        groups			= dict(
            group_parser( g )
            for g in group or GROUPS
        )
    assert groups and all( isinstance( k, str ) and len( v ) == 2 for k,v in groups.items() ), \
        f"Each group member specification must be a '<group>': (<needs>, <size>), not {type(next(groups.items()))}"

    group_threshold		= int( group_threshold ) if group_threshold else math.ceil( len( groups ) * GROUP_THRESHOLD_RATIO )

    cryptopaths			= list( cryptopaths_parser( cryptocurrency, edit=edit ))

    # If account details not provided in names, generate them.  If using_bip39 is specified, this is
    # where we use BIP-39 Seed generation to produce the wallet Seed, instead of SLIP-39 which uses
    # the entropy directly (optionally, with a passphrase, which is not typically done, and isn't
    # Trezor compatible).
    if not isinstance( names, dict ):
        assert not master_secret or not names or len( names ) == 1, \
            "Creating multiple account details from the same secret entropy doesn't make sense"
        names			= {
            name: create(
                name		= name,
                group_threshold	= group_threshold,
                groups		= groups,
                master_secret	= master_secret,
                passphrase	= passphrase.encode( 'UTF-8' ) if passphrase else b'',
                using_bip39	= using_bip39,  # Derive wallet Seed using BIP-39 Mnemonic + passphrase generation
                cryptopaths	= cryptopaths,
            )
            for name in names or [ "SLIP39" ]
        }

    if text and using_bip39:
        # Output the BIP-39 Mnemonic phrase we're using to generate the Seed as text.  We'll label
        # it as BIP-39, but it will just be ignored by standard SLIP-39 recovery attempts.
        print( f"Using BIP-39 Mnemonic: {produce_bip39( entropy=master_secret )}" )

    cover_text			= None

    # Generate each desired SLIP-39 Mnemonic cards PDF.  Supports --card (the default).  Remember
    # any deduced orientation and paper_format for below.
    results			= {}
    (pdf_paper,pdf_orient),pdf	= (None,None),None
    for name, details in names.items():
        # Get the first group of the accountgroups in details.accounts.
        accounts		= details.accounts
        assert accounts and accounts[0], \
            "At least one --cryptocurrency must be specified"
        for account in accounts[0]:
            log.warning( f"{account.crypto:6} {account.path:20}: {account.address}" )

        if text:
            # Output the SLIP-39 mnemonics as text, to stdout:
            #    name: <mnemonic>
            # or, if no name, just:
            #    <mnemonic>
            g_nam_max		= max( map( len, details.groups.keys() ))
            for g_name,(g_of,g_mnems) in details.groups.items():
                for i,mnem in enumerate( g_mnems ):
                    print( f"{name} {g_name:{g_nam_max}} {i+1}: {mnem}" )
        # Get the correct cover page format (if any), according to the details.using_bip39
        if cover_page:
            cover_text		= open(os.path.join(os.path.dirname(__file__), 'COVER.txt'), encoding='UTF-8').read()
            if using_bip39:
                cover_text     += open(os.path.join(os.path.dirname(__file__), 'COVER-BIP-39.txt', ), encoding='UTF-8').read()
            else:
                cover_text     += open(os.path.join(os.path.dirname(__file__), 'COVER-SLIP-39.txt'), encoding='UTF-8').read()

        # Unless no card_format (False) or paper wallet password specified, produce a PDF containing
        # the SLIP-39 mnemonic recovery cards; remember the deduced (<pdf_paper>,<pdf_orient>).  If
        # we're producing paper wallets, always force portrait orientation for the cards, to match.
        if card_format is not False or wallet_pwd is not None:
            (pdf_paper,pdf_orient),pdf,_ = produce_pdf(
                *details,
                card_format	= card_format or CARD,
                paper_format	= paper_format or PAPER,
                orientations	= ('portrait', ) if wallet_pwd is not None else None,
                cover_text	= cover_text,
                watermark	= watermark,
                double_sided	= double_sided,
            )

        now			= datetime.now()

        pdf_name		= ( filename or FILENAME_FORMAT ).format(
            name	= name,
            date	= datetime.strftime( now, '%Y-%m-%d' ),
            time	= datetime.strftime( now, '%H.%M.%S'),
            crypto	= accounts[0][0].crypto,
            path	= accounts[0][0].path,
            address	= accounts[0][0].address,
        )
        if not pdf_name.lower().endswith( '.pdf' ):
            pdf_name	       += '.pdf'

        if wallet_pwd is not None:
            # Deduce the paper wallet size and create a template.  All layouts are in specified in
            # inches; template dimensions are in mm.
            try:
                (wall_h,wall_w),wall_margin = WALLET_SIZES[wallet_format.lower() if wallet_format else WALLET]
            except KeyError:
                (wall_h,wall_w),wall_margin = ast.literal_eval( wallet_format )

            wall		= layout_wallet( Coordinate( y=wall_h, x=wall_w ), wall_margin )  # converts to mm
            wall_dim		= wall.mm()

            # Lay out wallets, always in Portrait orientation, defaulting to the Card paper_format
            # if it is a standard size (a str, not an (x,y) tuple), otherwise to "Letter" paper.  Printers may
            # have problems with a PDF mixing Landscape and Portrait, but do it if desired...
            if wallet_paper is None:
                wallet_paper	= paper_format if type(paper_format) is str else PAPER

            pdf.add_page( orientation='P', format=wallet_paper )
            page_margin_mm	= PAGE_MARGIN * MM_IN

            walls_pp,page_xy	= layout_components( pdf, comp_dim=wall_dim, page_margin_mm=page_margin_mm )
            wall_elements	= list( wall.elements() )
            if log.isEnabledFor( logging.DEBUG ):
                log.debug( f"Wallet elements: {json.dumps( wall_elements, indent=4)}" )
            wall_tpl		= fpdf.FlexTemplate( pdf, wall_elements )

            # Place each Paper Wallet adding pages as necessary (we already have the first fresh page).
            wall_n		= 0
            page_n		= 0
            for account_group in accounts:
                for c_n,account in enumerate( account_group ):
                    p,(offsetx,offsety) = page_xy( wall_n )
                    if p != page_n:
                        pdf.add_page( orientation='P', format=wallet_paper )
                        if double_sided is None or double_sided:
                            pdf.add_page( orientation='P', format=wallet_paper )
                        page_n	= p
                    try:
                        private_enc		= account.encrypted( wallet_pwd )
                    except NotImplementedError as exc:
                        log.exception( f"{account.crypto} doesn't support BIP-38 or JSON wallet encryption: {exc}" )
                        continue

                    wall_n     += 1

                    images			= os.path.dirname( __file__ )
                    wall_tpl['wallet-bg']	= os.path.join( images, 'paper-wallet-background.png' )
                    wall_tpl[f"crypto-f{c_n}"]	= account.crypto
                    wall_tpl[f"crypto-b{c_n}"]	= account.crypto

                    wall_tpl['center']		= os.path.join( images, account.crypto + '.png' )

                    wall_tpl['name-label']	= "Wallet:"
                    wall_tpl['name-bg']		= os.path.join( images, '1x1-ffffff54.png' )
                    wall_tpl['name']		= name

                    # wall_tpl['center-bg']	= os.path.join( images, 'guilloche-center.png' )

                    public_qr	= qrcode.QRCode(
                        version		= None,
                        error_correction = qrcode.constants.ERROR_CORRECT_M,
                        box_size	= 10,
                        border		= 1,
                    )
                    public_qr.add_data( account.address )
                    wall_tpl['address-l-bg']	= os.path.join( images, '1x1-ffffff54.png' )
                    wall_tpl['address-l']	= account.address
                    wall_tpl['address-r-bg']	= os.path.join( images, '1x1-ffffff54.png' )
                    wall_tpl['address-r']	= account.address

                    wall_tpl['address-qr-t']	= 'PUBLIC ADDRESS'
                    wall_tpl['address-qr-bg']	= os.path.join( images, '1x1-ffffff54.png' )
                    wall_tpl['address-qr']	= public_qr.make_image( back_color="transparent" ).get_image()
                    wall_tpl['address-qr-b']	= 'DEPOSIT/VERIFY'
                    wall_tpl['address-qr-r']	= account.path

                    private_qr	= qrcode.QRCode(
                        version		= None,
                        error_correction = qrcode.constants.ERROR_CORRECT_M,
                        box_size	= 10,
                        border		= 1,
                    )
                    private_qr.add_data( private_enc )

                    wall_tpl['private-bg']	= os.path.join( images, '1x1-ffffff54.png' )

                    # If not enough lines, will throw Exception, as it should!  We don't want to
                    # emit a Paper Wallet without the entire encrypted private key present.  This is
                    # primarily an issue for Ethereum encrypted JSON wallets, which are very large.
                    # As the vertical aspect ratio of the Paper Wallet increases (eg. for half- or
                    # third-page vs. quarter-page Paper Wallets), the line length increases, but the
                    # number of lines available decreases.  Estimate the number of characters on
                    # each line.
                    line_elm			= next( e for e in wall_elements if e['name'] == 'private-0' )
                    line_points			= ( line_elm['x2'] - line_elm['x1'] ) / MM_IN * PT_IN
                    line_fontsize		= line_elm['size']
                    line_chars			= int( line_points / line_fontsize / ( 5 / 8 ))  # Chars ~ 5/8 width vs. height
                    log.debug( f"Private key line length: {line_chars} chars" )
                    for ln,line in enumerate( chunker( private_enc, line_chars )):
                        wall_tpl[f"private-{ln}"] = line
                    wall_tpl['private-hint-t']	= 'PASSWORD HINT:'
                    wall_tpl['private-hint-bg']	= os.path.join( images, '1x1-ffffff54.png' )
                    wall_tpl['private-hint']	= wallet_pwd_hint
                    wall_tpl['private-qr-t']	= 'PRIVATE KEY'
                    wall_tpl['private-qr-bg']	= os.path.join( images, '1x1-ffffff54.png' )
                    wall_tpl['private-qr']	= private_qr.make_image( back_color="transparent" ).get_image()
                    wall_tpl['private-qr-b']	= 'SPEND'

                    wall_tpl.render( offsetx=offsetx, offsety=offsety )

        if json_pwd is not None:
            # If -j|--json supplied, also emit the encrypted JSON wallet.  This may be a *different*
            # password than the SLIP-39 master secret encryption passphrase.  It will be required to
            # decrypt and use the saved JSON wallet file, eg. to load a software Ethereum wallet.
            assert eth_account, \
                "The optional eth-account package is required to support output of encrypted JSON wallets\n" \
                "    python3 -m pip install eth-account"
            assert any( 'ETH' == crypto for crypto,*_ in cryptopaths ), \
                "--json is only valid if '--crypto ETH' wallets are specified"

            for eth in (
                account
                for group in accounts
                for account in group
                if account.crypto == 'ETH'
            ):
                json_str	= json.dumps( eth_account.Account.encrypt( eth.key, json_pwd ), indent=4 )
                json_name	= ( filename or FILENAME_FORMAT ).format(
                    name	= name or "SLIP39",
                    date	= datetime.strftime( now, '%Y-%m-%d' ),
                    time	= datetime.strftime( now, '%H.%M.%S'),
                    crypto	= eth.crypto,
                    path	= eth.path,
                    address	= eth.address,
                )
                if json_name.lower().endswith( '.pdf' ):
                    json_name	= json_name[:-4]
                json_name      += '.json'
                while os.path.exists( json_name ):
                    log.error( "ERROR: Will NOT overwrite {json_name}; adding '.new'!" )
                    json_name  += '.new'
                with open( json_name, 'w' ) as json_f:
                    json_f.write( json_str )
                log.warning( f"Wrote JSON {name or 'SLIP39'}'s encrypted ETH wallet {eth.address} derived at {eth.path} to: {json_name}" )

                if pdf:
                    # Add the encrypted JSON account recovery to the PDF also, if generated.  Last
                    # page, so no need to handle double_sided for subsequent output.
                    pdf.add_page()
                    margin_mm	= PAGE_MARGIN * MM_IN
                    pdf.set_margin( 1.0 * MM_IN )

                    col_width	= pdf.epw - 2 * margin_mm
                    pdf.set_font( FONTS['sans'], size=10 )
                    line_height	= pdf.font_size * 1.2
                    pdf.cell( col_width, line_height, json_name )
                    pdf.ln( line_height )

                    pdf.set_font( FONTS['sans'], size=9 )
                    line_height	= pdf.font_size * 1.1

                    for line in json_str.split( '\n' ):
                        pdf.cell( col_width, line_height, line )
                        pdf.ln( line_height )
                    pdf.image( qrcode.make( json_str ).get_image(), h=min(pdf.eph, pdf.epw)/2, w=min(pdf.eph, pdf.epw)/2 )

        if pdf:
            if filepath is not None:  # ''/True path implies current dir.
                if filepath is True:
                    filepath	= ''
                pdf_path		= os.path.join( filepath, pdf_name ) if filepath else pdf_name
                log.warning( f"Writing SLIP39{' backup for BIP-39' if using_bip39 else ''}-encoded wallet for {name!r} to: {pdf_path}" )
                pdf.output( pdf_path )
            if printer is not None:  # if True, uses "default" printer
                printer		= None if printer is True else printer
                log.warning( f"Printing SLIP39-encoded wallet for {name!r} to: {printer or '(default)'}: " )
                printer_output(
                    pdf.output(),
                    printer		= printer,
                    orientation		= pdf_orient,
                    paper_format	= pdf_paper,
                )

        results[pdf_name]	= details

    return results



import json
import logging
import pytest
import time

from fractions		import Fraction as F
from functools		import wraps
from collections	import defaultdict

from .util		import remainder_after, retry, timer, ordinal

log				= logging.getLogger( 'util_test' )


def test_remainder_after():
    assert list( remainder_after( [ .25, .25, .25, .25 ] )) == pytest.approx( [.75, 2/3, 1/2, 0] )
    # Now, the last segment requires double the remainder; so it extends -1x past the end...
    assert list( remainder_after( [ .25, .25, .25, .50 ] )) == pytest.approx( [.75, 2/3, 1/2, -1] )

    # Now, try some more complex Fractions, with differing desired total (eg. parts in 10,000), and
    # scale (multiplication factor required of each fraction to sum to desired total).

    parts			= [
        F( 1,  4 ),     # 3/12
        F( 1,  3 ),     # 4/12
        F( 1,  6 ),     # 2/12
        F( 1, 12 ),     # 1/12
        F( 1,  6 ),     # 2/12
    ]                  # 12/12
    assert sum( parts ) == 1
    assert list( remainder_after( parts )) == [
        F( 3, 4 ),
        F( 5, 9 ),
        F( 3, 5 ),
        F( 2, 3 ),
        0
    ]

    percs			= [
        f * 100 for f in parts
    ]
    assert sum( percs ) == 100
    assert list( remainder_after( percs, total=100 )) == [
        F( 75, 1 ), F( 500, 9 ), F( 60, 1 ), F( 200, 3 ), 0
    ]

    # Now, go straight from the ratio (0,1] to the parts in 10,000
    assert list( remainder_after( parts, scale=10000, total=10000 )) == [
        F( 7500, 1 ), F( 50000, 9 ), F( 6000, 1 ), F( 20000, 3 ), 0
    ]
    assert list( map( int, remainder_after( parts, scale=10000, total=10000 ))) == [
        7500, 5555, 6000, 6666, 0
    ]

    # If we're working entirely in Fractions and integers, make certain our results remain in
    # Fractions.  Lets make our proportions in range (0,1] as fixed-point x 2^16, and support
    # proportions that sum to something other than 1.  Since we are dealing in the *remainder*
    # (after sending the proportion to a recipient, in no case will a remainder ever equal 1, so we
    # don't have to worry about representing the range (0,1) in our outputs -- only (0,1].  Hence,
    # our denominator can be the full bit-width of the unsigned value.
    for factor in [ 1, 5, F( 99 / 100 ), F( 101, 100 )]:
        parts_factor		= [
            f * factor for f in parts
        ]
        rem_uint16		= list( remainder_after(
            parts_factor,
            scale	= ( 2 ** 16) / sum( parts_factor ),
            total	= ( 2 ** 16),
        ))
        assert rem_uint16 == [
            F(  49152, 1 ),
            F( 327680, 9 ),
            F( 196608, 5 ),
            F( 131072, 3 ),
            F(      0, 1 ),
        ]
        # And, sure enough, after correcting for our 2**16-1 scale, we get the originally calculated Fractions
        assert [
            f / ( 2 ** 16 ) for f in rem_uint16
        ] == [
            F( 3, 4 ),
            F( 5, 9 ),
            F( 3, 5 ),
            F( 2, 3 ),
            0
        ]


def exception_every( N=2, extra=0 ):
    if hasattr( extra, '__iter__' ):
        extra		= iter( extra )

    def decorator( func ):
        @wraps( func )
        def wrapper( *args, **kwds ):
            if wrapper.num > 0:
                wrapper.num    -= 1
                raise AssertionError( f"Fail for {wrapper.num} more times" )
            wrapper.nth	       += 1
            if wrapper.nth % N == 0:
                wrapper.num	= next( extra ) if hasattr( extra, '__next__' ) else extra
                raise AssertionError( f"Fails on {ordinal(wrapper.nth)} try, for {wrapper.num} more times" )
            return func( *args, **kwds )

        wrapper.nth		= 0
        wrapper.num		= 0

        return wrapper

    return decorator


def test_retry():
    """Lets target a backoff of 1.5x, and confirm at least 90% of that occurs.

    This is timing related, so we may have to make this an XFAIL test...

    """
    backoff			= 1.5

    @retry( 5, delay = .1, backoff = backoff, log_at = logging.DEBUG, exc_at=logging.INFO )
    @exception_every( 2, extra=range( 10 ))
    def truthy():
        truthy.i	       += 1
        return truthy.i

    truthy.i			= 0

    results			= {}
    dur				= 0.05
    now = then 			= timer()
    for i in range( 98 ):
        ela			= i * dur
        time.sleep( max( 0, ela - (now - then )))
        results[f"{ela:6.4f}s"] = truthy()
        now			= timer()

    log.debug( json.dumps( results, indent=4 ))

    counts			= defaultdict( int )

    for v in results.values():
        counts[v]	       += 1

    log.debug( json.dumps( counts, indent=4 ))

    for k in counts.keys():
        if k-1 in counts:
            log.info( f" {k} vs {k-1}: {counts[k]/counts[k-1]:7.3f}x" )
            # Expected exponential backoff w/ increasing failures
            assert counts[k] > counts[k-1] * backoff * 90/100
